from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    CollapseWidget,
    MarkdownWidget,
    NotificationWidget,
)


class IOUDistribution(DetectionVisMetric):
    MARKDOWN_LOCALIZATION_ACCURACY = "localization_accuracy"
    MARKDOWN_IOU_DISTRIBUTION = "iou_distribution"
    NOTIFICATION = "iou_distribution"
    COLLAPSE = "iou_distribution"
    CHART = "iou_distribution"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.md_title = "Localization Accuracy (IoU)"

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_localization_accuracy
        text = text.format(self.vis_texts.definitions.iou_score)
        return MarkdownWidget(self.MARKDOWN_LOCALIZATION_ACCURACY, self.md_title, text)

    @property
    def md_iou_distribution(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_iou_distribution
        text = text.format(self.vis_texts.definitions.iou_score)
        return MarkdownWidget(self.MARKDOWN_IOU_DISTRIBUTION, self.md_title, text)

    @property
    def notification(self) -> NotificationWidget:
        title, _ = self.vis_texts.notification_avg_iou.values()
        return NotificationWidget(
            self.NOTIFICATION,
            title.format(self.eval_result.mp.base_metrics()["iou"].round(2)),
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(self.CHART, self._get_figure())

    @property
    def collapse(self) -> CollapseWidget:
        md1 = MarkdownWidget(
            "iou_calculation",
            "How IoU is calculated?",
            self.vis_texts.markdown_iou_calculation,
        )
        md2 = MarkdownWidget(
            "what_is_pr_curve",
            "How the PR curve is built?",
            self.vis_texts.markdown_what_is_pr_curve.format(
                self.vis_texts.definitions.confidence_score,
                self.vis_texts.definitions.true_positives,
                self.vis_texts.definitions.false_positives,
            ),
        )
        return CollapseWidget([md1, md2])

    def _get_figure(self):  # -> go.Figure:
        import plotly.graph_objects as go  # pylint: disable=import-error

        fig = go.Figure()
        nbins = 40
        fig.add_trace(go.Histogram(x=self.eval_result.mp.ious, nbinsx=nbins))
        fig.update_layout(
            xaxis_title="IoU",
            yaxis_title="Count",
            width=600,
            height=500,
        )

        # Add annotation for mean IoU as vertical line
        mean_iou = self.eval_result.mp.ious.mean()
        y1 = len(self.eval_result.mp.ious) // nbins
        fig.add_shape(
            type="line",
            x0=mean_iou,
            x1=mean_iou,
            y0=0,
            y1=y1,
            line=dict(color="orange", width=2, dash="dash"),
        )
        fig.update_traces(hovertemplate="IoU: %{x:.2f}<br>Count: %{y}<extra></extra>")
        fig.add_annotation(x=mean_iou, y=y1, text=f"Mean IoU: {mean_iou:.2f}", showarrow=False)
        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig



from typing import List, Tuple

from supervisely.annotation.annotation import Annotation
from supervisely.api.api import Api
from supervisely.api.image_api import ImageInfo
from supervisely.api.module_api import ApiField
from supervisely.api.project_api import ProjectInfo
from supervisely.nn.benchmark.base_evaluator import BaseEvalResult
from supervisely.nn.benchmark.cv_tasks import CVTask
from supervisely.nn.benchmark.visualization.renderer import Renderer
from supervisely.nn.benchmark.visualization.widgets import GalleryWidget
from supervisely.project.project_meta import ProjectMeta
from supervisely.task.progress import tqdm_sly


class MatchedPairData:
    def __init__(
        self,
        gt_image_info: ImageInfo = None,
        pred_image_info: ImageInfo = None,
        diff_image_info: ImageInfo = None,
        gt_annotation: Annotation = None,
        pred_annotation: Annotation = None,
        diff_annotation: Annotation = None,
    ):
        self.gt_image_info = gt_image_info
        self.pred_image_info = pred_image_info
        self.diff_image_info = diff_image_info
        self.gt_annotation = gt_annotation
        self.pred_annotation = pred_annotation
        self.diff_annotation = diff_annotation


class BaseVisMetrics:

    def __init__(
        self,
        vis_texts,
        eval_results: List[BaseEvalResult],
        explore_modal_table: GalleryWidget = None,
        diff_modal_table: GalleryWidget = None,
    ) -> None:
        self.vis_texts = vis_texts
        self.eval_results = eval_results
        self.explore_modal_table = explore_modal_table
        self.diff_modal_table = diff_modal_table
        self.clickable = False


class BaseVisMetric(BaseVisMetrics):
    def __init__(
        self,
        vis_texts,
        eval_result: BaseEvalResult,
        explore_modal_table: GalleryWidget = None,
        diff_modal_table: GalleryWidget = None,
    ) -> None:
        super().__init__(vis_texts, [eval_result], explore_modal_table, diff_modal_table)
        self.eval_result = eval_result


class BaseVisualizer:
    cv_task = None
    report_name = "Model Evaluation Report.lnk"

    def __init__(
        self,
        api: Api,
        eval_results: List[BaseEvalResult],
        workdir="./visualizations",
        progress=None,
    ):
        self.api = api
        self.workdir = workdir
        self.eval_result = eval_results[0]  # for evaluation
        self.eval_results = eval_results  # for comparison

        self.renderer = None
        self.gt_project_info = None
        self.gt_project_meta = None
        self.gt_dataset_infos = None
        self.pbar = progress or tqdm_sly
        self.ann_opacity = 0.4

        with self.pbar(message="Fetching project and dataset infos", total=len(eval_results)) as p:
            for eval_result in self.eval_results:
                self._get_eval_project_infos(eval_result)
                p.update(1)

    def _get_eval_project_infos(self, eval_result):
        # get project infos
        if self.gt_project_info is None:
            self.gt_project_info = self.api.project.get_info_by_id(eval_result.gt_project_id)
        eval_result.gt_project_info = self.gt_project_info
        eval_result.pred_project_info = self.api.project.get_info_by_id(eval_result.pred_project_id)

        # get project metas
        if self.gt_project_meta is None:
            self.gt_project_meta = ProjectMeta.from_json(
                self.api.project.get_meta(eval_result.gt_project_id)
            )
        eval_result.gt_project_meta = self.gt_project_meta
        eval_result.pred_project_meta = ProjectMeta.from_json(
            self.api.project.get_meta(eval_result.pred_project_id)
        )

        # get dataset infos
        filters = None
        if eval_result.gt_dataset_ids is not None:
            filters = [
                {
                    ApiField.FIELD: ApiField.ID,
                    ApiField.OPERATOR: "in",
                    ApiField.VALUE: eval_result.gt_dataset_ids,
                }
            ]
        if self.gt_dataset_infos is None:
            self.gt_dataset_infos = self.api.dataset.get_list(
                eval_result.gt_project_id,
                filters=filters,
                recursive=True,
            )
        eval_result.gt_dataset_infos = self.gt_dataset_infos
        filters = [
            {
                ApiField.FIELD: ApiField.NAME,
                ApiField.OPERATOR: "in",
                ApiField.VALUE: [ds.name for ds in self.gt_dataset_infos],
            }
        ]
        eval_result.pred_dataset_infos = self.api.dataset.get_list(
            eval_result.pred_project_id, filters=filters, recursive=True
        )

        # get train task info
        train_info = eval_result.train_info
        if train_info:
            train_task_id = train_info.get("app_session_id")
            if train_task_id:
                eval_result.task_info = self.api.task.get_info_by_id(int(train_task_id))

    def visualize(self):
        if self.renderer is None:
            layout = self._create_layout()
            self.renderer = Renderer(layout, self.workdir, report_name=self.report_name)
        return self.renderer.visualize()

    def upload_results(self, team_id: int, remote_dir: str, progress=None):
        if self.renderer is None:
            raise RuntimeError("Visualize first")
        return self.renderer.upload_results(self.api, team_id, remote_dir, progress)

    def _create_layout(self):
        raise NotImplementedError("Implement this method in a subclass")

    def _get_or_create_diff_project(self) -> Tuple[ProjectInfo, List, bool]:
        """
        Get or create a project for diff visualizations.
        Dataset hierarchy is copied from the prediction project.
        """

        pred_ds_id_to_diff_ds_info = {}
        diff_ds_infos = []

        def _get_or_create_diff_dataset(pred_dataset_id, pred_datasets):
            if pred_dataset_id in pred_ds_id_to_diff_ds_info:
                return pred_ds_id_to_diff_ds_info[pred_dataset_id]
            pred_dataset = pred_datasets[pred_dataset_id]
            if pred_dataset.parent_id is None:
                diff_dataset = self.api.dataset.create(project_info.id, pred_dataset.name)
            else:
                parent_dataset = _get_or_create_diff_dataset(pred_dataset.parent_id, pred_datasets)
                diff_dataset = self.api.dataset.create(
                    project_info.id,
                    pred_dataset.name,
                    parent_id=parent_dataset.id,
                )
            pred_ds_id_to_diff_ds_info[pred_dataset_id] = diff_dataset
            diff_ds_infos.append(diff_dataset)
            return diff_dataset

        is_existed = False
        project_name = self._generate_diff_project_name(self.eval_result.pred_project_info.name)
        workspace_id = self.eval_result.pred_project_info.workspace_id
        project_info = self.api.project.create(
            workspace_id, project_name, change_name_if_conflict=True
        )
        pred_datasets = {ds.id: ds for ds in self.eval_result.pred_dataset_infos}
        for dataset in pred_datasets:
            _get_or_create_diff_dataset(dataset, pred_datasets)
        return project_info, diff_ds_infos, is_existed

    def _generate_diff_project_name(self, pred_project_name):
        return "[diff]: " + pred_project_name

    def _create_explore_modal_table(
        self, columns_number=3, click_gallery_id=None, hover_text=None
    ) -> GalleryWidget:
        gallery = GalleryWidget(
            "all_predictions_modal_gallery",
            is_modal=True,
            columns_number=columns_number,
            click_gallery_id=click_gallery_id,
            opacity=self.ann_opacity,
        )
        gallery.set_project_meta(self.eval_results[0].filtered_project_meta)
        if hover_text:
            gallery.add_image_left_header(hover_text)
        return gallery

    def _create_diff_modal_table(self, columns_number=3) -> GalleryWidget:
        gallery = GalleryWidget(
            "diff_predictions_modal_gallery",
            is_modal=True,
            columns_number=columns_number,
            opacity=self.ann_opacity,
        )
        gallery.set_project_meta(self.eval_results[0].filtered_project_meta)
        return gallery

    def _get_filtered_project_meta(self, eval_result) -> ProjectMeta:
        remove_classes = []
        meta = eval_result.pred_project_meta.clone()
        if eval_result.classes_whitelist:
            for obj_class in meta.obj_classes:
                if obj_class.name not in eval_result.classes_whitelist:
                    remove_classes.append(obj_class.name)
            if remove_classes:
                meta = meta.delete_obj_classes(remove_classes)
        return meta

    def _update_match_data(
        self,
        gt_image_id: int,
        gt_image_info: ImageInfo = None,
        pred_image_info: ImageInfo = None,
        diff_image_info: ImageInfo = None,
        gt_annotation: Annotation = None,
        pred_annotation: Annotation = None,
        diff_annotation: Annotation = None,
    ):
        match_data = self.eval_result.matched_pair_data.get(gt_image_id, None)
        if match_data is None:
            self.eval_result.matched_pair_data[gt_image_id] = MatchedPairData(
                gt_image_info=gt_image_info,
                pred_image_info=pred_image_info,
                diff_image_info=diff_image_info,
                gt_annotation=gt_annotation,
                pred_annotation=pred_annotation,
                diff_annotation=diff_annotation,
            )
        else:
            for attr, value in {
                "gt_image_info": gt_image_info,
                "pred_image_info": pred_image_info,
                "diff_image_info": diff_image_info,
                "gt_annotation": gt_annotation,
                "pred_annotation": pred_annotation,
                "diff_annotation": diff_annotation,
            }.items():
                if value is not None:
                    setattr(match_data, attr, value)



import numpy as np
import torch  # pylint: disable=import-error
from PIL import Image


def extract_image_patch(image, bbox, patch_shape=None):
    """Extract image patch from bounding box.

    Parameters
    ----------
    image : ndarray
        The full image.
    bbox : array_like
        The bounding box in format (x, y, width, height).
    patch_shape : Optional[array_like]
        This parameter can be used to enforce a desired patch shape
        (height, width). First, the `bbox` is adapted to the aspect ratio
        of the patch shape, then it is clipped at the image boundaries.
        If None, the shape is computed from :arg:`bbox`.

    Returns
    -------
    ndarray | NoneType
        An image patch showing the :arg:`bbox`, optionally reshaped to
        :arg:`patch_shape`.
        Returns None if the bounding box is empty or fully outside of the image
        boundaries.

    """
    bbox = np.array(bbox)
    if patch_shape is not None:
        # correct aspect ratio to patch shape
        target_aspect = float(patch_shape[1]) / patch_shape[0]
        new_width = target_aspect * bbox[3]
        bbox[0] -= (new_width - bbox[2]) / 2
        bbox[2] = new_width

    # convert to top left, bottom right
    bbox[2:] += bbox[:2]
    bbox = bbox.astype(np.int)

    # clip at image boundaries
    bbox[:2] = np.maximum(0, bbox[:2])
    bbox[2:] = np.minimum(np.asarray(image.shape[:2][::-1]) - 1, bbox[2:])
    if np.any(bbox[:2] >= bbox[2:]):
        return None
    sx, sy, ex, ey = bbox
    image = image[sy:ey, sx:ex]
    # image = cv2.resize(image, tuple(patch_shape[::-1]))
    return image


class ImageEncoder(object):

    def __init__(self, model, transform, device):

        self.model = model
        self.transform = transform
        self.device = str(device)

    def __call__(self, data_x, batch_size=32):
        out = []

        for patch in range(len(data_x)):
            if self.device == "cpu":
                img = self.transform(Image.fromarray(data_x[patch]))
            else:
                img = self.transform(Image.fromarray(data_x[patch])).cuda()
            out.append(img)

        features = self.model.encode_image(torch.stack(out)).cpu().detach().numpy()
        return features


def create_box_encoder(model, transform, batch_size=32, device="cpu"):
    image_encoder = ImageEncoder(model, transform, device)

    def encoder(image, boxes):
        image_patches = []
        for box in boxes:
            # print("extracting box {} from image {}".format(box, image.shape))
            patch = extract_image_patch(image, box)
            if patch is None:
                print("WARNING: Failed to extract image patch: %s." % str(box))
                patch = np.random.uniform(0.0, 255.0, image.shape).astype(np.uint8)
            image_patches.append(patch)
        # image_patches = np.array(image_patches)
        return image_encoder(image_patches, batch_size)

    return encoder



from types import SimpleNamespace

docs_url = "https://docs.supervisely.com/neural-networks/model-evaluation-benchmark/"

definitions = SimpleNamespace(
    true_positives="True Positives (TP): These are correctly detected objects. For a prediction to be counted as a true positive, the predicted bounding box must align with a ground truth bounding box with an Intersection over Union (IoU) of 0.5 or more, and the object must be correctly classified",
    false_positives="False Positives (FP): These are incorrect detections made by the model. They occur when the model predicts a bounding box that either does not overlap sufficiently with any ground truth box (IoU less than 0.5) or incorrectly classifies the object within the bounding box. For example, the model detects a car in the image, but there is no car in the ground truth.",
    false_negatives="False Negatives (FN): These are the missed detections. They occur when an actual object in the ground truth is not detected by the model, meaning there is no predicted bounding box with an IoU of 0.5 or more for this object. For example, there is a car in the image, but the model fails to detect it.",
    confidence_threshold="Confidence threshold is a hyperparameter used to filter out predictions that the model is not confident in. By setting a higher confidence threshold, you ensure that only the most certain predictions are considered, thereby reducing the number of false predictions. This helps to control the trade-off between precision and recall in the model's output.",
    confidence_score="The confidence score, also known as probability score, quantifies how confident the model is that its prediction is correct. It is a numerical value between 0 and 1, generated by the model for each bounding box, that represents the likelihood that a predicted bounding box contains an object of a particular class.",
    f1_score="F1-score is a useful metric that combines both precision and recall into a single measure. As the harmonic mean of precision and recall, the f1-score provides a balanced representation of both metrics in one value. F1-score ranges from 0 to 1, with a higher score indicating better model performance. It is calculated as 2 * (precision * recall) / (precision + recall).",
    average_precision="Average precision (AP) is computed as the area under the precision-recall curve. It measures the precision of the model at different recall levels and provides a single number that summarizes the trade-off between precision and recall for a given class.",
    about_pr_tradeoffs="A system with high recall but low precision returns many results, but most of its predictions are incorrect or redundant (false positive). A system with high precision but low recall is just the opposite, returning very few results, most of its predictions are correct. An ideal system with high precision and high recall will return many results, with all results predicted correctly.",
    iou_score="IoU score is a measure of overlap between predicted bounding box and ground truth bounding box. A higher IoU score indicates better alignment between the predicted and ground truth bounding boxes.",
    iou_threshold="The IoU threshold is a predefined value (set to 0.5 in many benchmarks) that determines the minimum acceptable IoU score for a predicted bounding box to be considered a correct prediction. When the IoU of a predicted box and actual box is higher than this IoU threshold, the prediction is considered correct. Some metrics will evaluate the model with different IoU thresholds to provide more insights about the model's performance.",
)

# <i class="zmdi zmdi-check-circle" style="color: #13ce66; margin-right: 5px"></i>
clickable_label = """
> <span style="color: #5a6772">
>     Click on the chart to explore corresponding images.
> </span>
"""

markdown_header = """
<h1>{}</h1>

<div class="model-info-block">
    <div>Created by <b>{}</b></div>
    <div><i class="zmdi zmdi-calendar-alt"></i><span>{}</span></div>
</div>
"""

markdown_common_overview = """
- **Models**: {}
- **Evaluation Dataset**: <a href="/projects/{}/datasets" target="_blank">{}</a>
- **Task type**: {}
"""

markdown_overview_info = """
<h3>{}</h3>
- **Model**: {}
- **Checkpoint**: {}
- **Architecture**: {}
- **Runtime**: {}
- **Checkpoint file**: <a class="checkpoint-url" href="{}" target="_blank">{}</a>
- **Evaluation Report**: <a href="{}" target="_blank">View Report</a>

"""

markdown_overview = """
- **Model**: {}
- **Checkpoint**: {}
- **Architecture**: {}
- **Task type**: {}
- **Runtime**: {}
- **Checkpoint file**: <a class="checkpoint-url" href="{}" target="_blank">{}</a>
- **Ground Truth project**: <a href="/projects/{}/datasets" target="_blank">{}</a>, {}{}
{}

Learn more about Model Benchmark, implementation details, and how to use the charts in our <a href="{}" target="_blank">Technical Report</a>.
"""
# - **Model**: {}
# - **Training dataset (?)**: COCO 2017 train
# - **Model classes (?)**: (80): a, b, c, … (collapse)
# - **Model weights (?)**: [/path/to/yolov8l.pt]()
# - **License (?)**: AGPL-3.0

markdown_key_metrics = """## Key Metrics

Here, we comprehensively assess the model's performance by presenting a broad set of metrics, including mAP (mean Average Precision), Precision, Recall, IoU (Intersection over Union), Classification Accuracy, Calibration Score, and Inference Speed.

- **Mean Average Precision (mAP)**: A comprehensive metric of detection performance. mAP calculates the <abbr title="{}">average precision</abbr> across all classes at different levels of <abbr title="{}">IoU thresholds</abbr> and precision-recall trade-offs. In other words, it evaluates the performance of a model by considering its ability to detect and localize objects accurately across multiple IoU thresholds and object categories.
- **Precision**: Precision indicates how often the model's predictions are actually correct when it predicts an object. This calculates the ratio of correct detections to the total number of detections made by the model.
- **Recall**: Recall measures the model's ability to find all relevant objects in a dataset. This calculates the ratio of correct detections to the total number of instances in a dataset.
- **Intersection over Union (IoU)**: IoU measures how closely predicted bounding boxes match the actual (ground truth) bounding boxes. It is calculated as the area of overlap between the predicted bounding box and the ground truth bounding box, divided by the area of union of these bounding boxes.
- **Classification Accuracy**: We additionally measure the classification accuracy of an object detection model. This metric represents the percentage of correctly labeled instances among all correctly localized bounding boxes (where the IoU for each box is greater than 0.5, regardless of class).
- **Calibration Score**: This score represents the consistency of predicted probabilities (or <abbr title="{}">confidence scores</abbr>) made by the model. We evaluate how well predicted probabilities align with actual outcomes. A well-calibrated model means that when it predicts an object with, say, 80% confidence, approximately 80% of those predictions should actually be correct.
- **Inference Speed**: The number of frames per second (FPS) the model can process, measured with a batch size of 1. The inference speed is important in applications, where real-time object detection is required. Additionally, slower models pour more GPU resources, so their inference cost is higher.
"""

markdown_explorer = """## Explore Predictions
In this section you can visually assess the model performance through examples. This helps users better understand model capabilities and limitations, giving an intuitive grasp of prediction quality in different scenarios.

> Click on the image to view the **Ground Truth**, **Prediction**, and **Difference** annotations side-by-side. 

> Filtering options allow you to adjust the confidence threshold (only for predictions) and the model's false outcomes (only for differences). Differences are calculated only for the optimal confidence threshold, allowing you to focus on the most accurate predictions made by the model.
"""

markdown_explore_difference = """## Explore Predictions

In this section, you can explore predictions made by different models side-by-side. This helps you to understand the differences in predictions made by each model, and to identify which model performs better in different scenarios.


> Click on the image to view the **Ground Truth**, and **Prediction** annotations side-by-side.
"""
### Difference in Predictions

markdown_explore_same_errors = """
### Same Errors

This section helps you to identify samples where all models made the same errors. It is useful for understanding the limitations of the models and the common challenges they face.

> Click on the image to view the **Ground Truth**, and **Prediction** annotations side-by-side.
"""

markdown_predictions_gallery = """

"""
# You can choose one of the sorting method:

# - **Auto**: The algorithm is trying to gather a diverse set of images that illustrate the model's performance across various scenarios.
# - **Least accurate**: Displays images where the model made more errors.
# - **Most accurate**: Displays images where the model made fewer or no errors.
# - **Dataset order**: Displays images in the original order of the dataset.
# """

markdown_predictions_table = """### Prediction details for every image

The table helps you in finding samples with specific cases of interest. You can sort by parameters such as the number of predictions, or specific a metric, e.g, recall, then click on a row to view this image and predictions.

**Example**: you can sort by **FN** (False Negatives) in descending order to identify samples where the model failed to detect many objects.


> Click on the row to view the image with **Ground Truth**, **Prediction**, or the **Difference** annotations.
"""

markdown_what_is = """
"""

markdown_experts = """
"""

markdown_how_to_use = """
"""

markdown_outcome_counts = (
    """## Outcome Counts

This chart is used to evaluate the overall model performance by breaking down all predictions into <abbr title="{}">True Positives</abbr> (TP), <abbr title="{}">False Positives</abbr> (FP), and <abbr title="{}">False Negatives</abbr> (FN). This helps to visually assess the type of errors the model often encounters.

"""
    + clickable_label
)

markdown_outcome_counts_diff = (
    """### Outcome Counts Differences

This chart compares the outcomes between different models. It helps identify where the models agree and disagree in their predictions. The 'Common' bar represents cases where the models made identical predictions, regardless of whether those predictions were correct or incorrect. The individual model bars show instances where that particular model made unique predictions that differed from other models.

"""
    + clickable_label
)

markdown_precision_per_class_title = """### Precision by Class"""

markdown_recall_per_class_title = """### Recall by Class"""

markdown_f1_per_class_title = """### F1-score by Class"""

markdown_R = """## Recall

This section measures the ability of the model to detect **all relevant instances in the dataset**. In other words, it answers the question: "Of all instances in the dataset, how many of them is the model managed to find out?"

To measure this, we calculate **Recall**. Recall counts errors, when the model does not detect an object that actually is present in a dataset and should be detected. Recall is calculated as the portion of correct predictions (true positives) over all instances in the dataset (true positives + false negatives).
"""

notification_recall = {
    "title": "Recall = {}",
    "description": "The model correctly found <b>{} of {}</b> total instances in the dataset.",
}

markdown_R_perclass = (
    """### Per-class Recall

This chart further analyzes Recall, breaking it down to each class in separate.

Since the overall recall is calculated as an average across all classes, we provide a chart showing the recall for each individual class. This illustrates how much each class contributes to the overall recall.

_Bars in the chart are sorted by <abbr title="{}">F1-score</abbr> to keep a unified order of classes between different charts._

"""
    + clickable_label
)


markdown_P = """## Precision

This section measures the accuracy of all predictions made by the model. In other words, it answers the question: "Of all predictions made by the model, how many of them are actually correct?".

To measure this, we calculate **Precision**. Precision counts errors, when the model predicts an object (bounding box), but the image has no objects of the predicted class in this place. Precision is calculated as a portion of correct predictions (true positives) over all model's predictions (true positives + false positives).
"""

notification_precision = {
    "title": "Precision = {}",
    "description": "The model correctly predicted <b>{} of {}</b> predictions made by the model in total.",
}

markdown_P_perclass = (
    """### Per-class Precision

This chart further analyzes Precision, breaking it down to each class in separate.

Since the overall precision is computed as an average across all classes, we provide a chart showing the precision for each class individually. This illustrates how much each class contributes to the overall precision.

_Bars in the chart are sorted by <abbr title="{}">F1-score</abbr> to keep a unified order of classes between different charts._

"""
    + clickable_label
)


markdown_PR = (
    """## Recall vs. Precision

This section compares Precision and Recall in one graph, identifying **imbalance** between these two.

_Bars in the chart are sorted by <abbr title="{}">F1-score</abbr> to keep a unified order of classes between different charts._

"""
    + clickable_label
)

markdown_PRF1 = """## Precision, Recall, F1-score

This section compares Precision, Recall, and F1-score metrics. The first graph identifies imbalance between these metrics. On the next charts, you can see each metric separately, broken down by class.

_Bars in the chart are sorted by <abbr title="{}">F1-score</abbr> to keep a unified order of classes between different charts._
"""

markdown_pr_curve = """## mAP & Precision-Recall Curve

Precision-Recall curve is an overall performance indicator. It helps to visually assess both precision and recall for all predictions made by the model on the whole dataset. This gives you an understanding of how precision changes as you attempt to increase recall, providing a view of **trade-offs between precision and recall** <abbr title="{}">(?)</abbr>. Ideally, a high-quality model will maintain strong precision as recall increases. This means that as you move from left to right on the curve, there should not be a significant drop in precision. Such a model is capable of finding many relevant instances, maintaining a high level of precision.
"""

markdown_trade_offs = """- A system with high recall but low precision generates many results, but most of its predictions are incorrect or redundant (false positives).

- Conversely, a system with high precision but low recall produces very few results, but most of its predictions are accurate.

- The ideal system achieves both high precision and high recall, meaning it returns many results with a high accuracy rate.
"""

markdown_what_is_pr_curve = """1. **Sort predictions**: Arrange all bounding box predictions by their <abbr title="{}">confidence scores</abbr> in descending order.

2. **Classify outcomes**: For each prediction, determine if it is a <abbr title="{}">true positive</abbr> (TP) or a <abbr title="{}">false positive</abbr> (FP) and record these classifications in a table.

3. **Calculate cumulative metrics**: As you move through each prediction, calculate the cumulative precision and recall. Add these values to the table.

4. **Plot points**: Each row in the table now represents a point on a graph, with cumulative recall on the x-axis and cumulative precision on the y-axis. Initially, this creates a zig-zag line because of variations between predictions.

5. **Smooth the curve**: The true PR curve is derived by plotting only the maximum precision value for each recall level across all thresholds. This means you connect only the highest points of precision for each segment of recall, smoothing out the zig-zags and forming a curve that typically slopes downward as recall increases.
"""


notification_ap = {
    "title": "mAP = {}",
    "description": "",
}

markdown_pr_by_class = (
    """### Precision-Recall Curve by Class

In this plot, you can evaluate PR curve for each class individually.

"""
    + clickable_label
)

markdown_confusion_matrix = (
    """## Confusion Matrix

Confusion matrix helps to find the number of confusions between different classes made by the model.
Each row of the matrix represents the instances in a ground truth class, while each column represents the instances in a predicted class.
The diagonal elements represent the number of correct predictions for each class (True Positives), and the off-diagonal elements show misclassifications.

"""
    + clickable_label
)


markdown_frequently_confused = (
    """### Frequently Confused Classes

This chart displays the most frequently confused pairs of classes. In general, it finds out which classes visually seem very similar to the model.

The chart calculates the **probability of confusion** between different pairs of classes. For instance, if the probability of confusion for the pair "{} - {}" is {}, this means that when the model predicts either "{}" or "{}", there is a {}% chance that the model might mistakenly predict one instead of the other.

The measure is class-symmetric, meaning that the probability of confusing a {} with a {} is equal to the probability of confusing a {} with a {}.

"""
    + clickable_label
)


markdown_localization_accuracy = """## Localization Accuracy (IoU)

This section measures how closely predicted bounding boxes generated by the model are aligned with the actual (ground truth) bounding boxes.
"""

markdown_iou_calculation = """<img src='https://github.com/dataset-ninja/model-benchmark-template/assets/78355358/8d7c63d0-2f3b-4f3f-9fd8-c6383a4bfba4' alt='alt text' width='300' />

To measure it, we calculate the <b>Intersection over Union (IoU)</b>. Intuitively, the higher the IoU, the closer two bounding boxes are. IoU is calculated by dividing the <b>area of overlap</b> between the predicted bounding box and the ground truth bounding box by the <b>area of union</b> of these two boxes.
"""

markdown_iou_distribution = """### IoU Distribution

This histogram represents the distribution of <abbr title="{}">IoU scores</abbr> among all predictions. This gives you a sense of how well the model aligns bounding boxes. Ideally, the rightmost bars (from 0.9 to 1.0 IoU) should be much higher than others.
"""


notification_avg_iou = {
    "title": "Avg. IoU = {}",
    "description": "",
}

markdown_calibration_score_1 = """## Calibration Score

This section analyzes <abbr title="{}">confidence scores</abbr> (or predicted probabilities) that the model generates for every predicted bounding box.
"""

markdown_what_is_calibration = """In some applications, it's crucial for a model not only to make accurate predictions but also to provide reliable **confidence levels**. A well-calibrated model aligns its confidence scores with the actual likelihood of predictions being correct. For example, if a model claims 90% confidence for predictions but they are correct only half the time, it is **overconfident**. Conversely, **underconfidence** occurs when a model assigns lower confidence scores than the actual likelihood of its predictions. In the context of autonomous driving, this might cause a vehicle to brake or slow down too frequently, reducing travel efficiency and potentially causing traffic issues."""
markdown_calibration_score_2 = """To evaluate the calibration, we draw a <b>Reliability Diagram</b> and calculate <b>Expected Calibration Error</b> (ECE)."""

markdown_reliability_diagram = """## Reliability Diagram

Reliability diagram, also known as a Calibration curve, helps in understanding whether the confidence scores of detections accurately represent the true probability of a correct detection. A well-calibrated model means that when it predicts a detection with, say, 80% confidence, approximately 80% of those predictions should actually be correct.
"""

markdown_calibration_curve_interpretation = """
1. **The curve is above the Ideal Line (Underconfidence):** If the calibration curve is consistently above the ideal line, this indicates underconfidence. The model's predictions are more correct than the confidence scores suggest. For example, if the model predicts a detection with 70% confidence but, empirically, 90% of such detections are correct, the model is underconfident.
2. **The curve is below the Ideal Line (Overconfidence):** If the calibration curve is below the ideal line, the model exhibits overconfidence. This means it is too sure of its predictions. For instance, if the model predicts with 80% confidence but only 60% of these predictions are correct, it is overconfident.

To quantify the calibration, we calculate **Expected Calibration Error (ECE).** Intuitively, ECE can be viewed as a deviation of the model's calibration curve from the diagonal line, that corresponds to a perfectly calibrated model. When ECE is high, we can not trust predicted probabilities so much.

**Note:** ECE is a measure of **error**. The lower the ECE, the better the calibration. A perfectly calibrated model has an ECE of 0.
"""

notification_ece = {
    "title": "Expected Calibration Error (ECE) = {}",
    "description": "",
}


markdown_confidence_score_1 = """## Confidence Score Profile

This section is going deeper in analyzing confidence scores. It gives you an intuition about how these scores are distributed and helps to find the best <abbr title="{}">confidence threshold</abbr> suitable for your task or application.
"""

markdown_confidence_score_2 = """This chart provides a comprehensive view about predicted confidence scores. It is used to determine an **optimal confidence threshold** based on your requirements.

The plot shows you what the metrics will be if you choose a specific confidence threshold. For example, if you set the threshold to 0.32, you can see on the plot what the precision, recall and f1-score will be for this threshold.
"""

markdown_plot_confidence_profile = """
First, we sort all predictions by confidence scores from highest to lowest. As we iterate over each prediction we calculate the cumulative precision, recall and f1-score so far. Each prediction is plotted as a point on a graph, with a confidence score on the x-axis and one of three metrics on the y-axis (precision, recall, f1-score).
"""

markdown_calibration_score_3 = """**How to find an optimal threshold:** you can find the maximum of the f1-score line on the plot, and the confidence score (X-axis) under this maximum corresponds to F1-optimal confidence threshold. This threshold ensures the balance between precision and recall. You can select a threshold according to your desired trade-offs."""

notification_f1 = {
    "title": "F1-optimal confidence threshold = {}",
    "description": "",
}

markdown_f1_at_ious = """### Confidence Profile at Different IoU thresholds

This chart breaks down the Confidence Profile into multiple curves, each for one <abbr title="{}">IoU threshold</abbr>. In this way you can understand how the f1-optimal confidence threshold changes with various IoU thresholds. Higher IoU thresholds mean that the model should align bounding boxes more accurately. This chart helps to find the optimal confidence threshold for different levels of localization accuracy.
"""
markdown_confidence_distribution = """### Confidence Distribution

This graph helps to assess whether high confidence scores correlate with correct predictions (<abbr title="{}">true positives</abbr>) and the low confidence scores correlate with incorrect ones (<abbr title="{}">false positives</abbr>). It consists of two histograms, one for true positive predictions filled with green, and one for false positives filled with red.

Additionally, it provides a view of how predicted probabilities are distributed. Whether the model skews probabilities to lower or higher values, leading to imbalance?

Ideally, the green histogram (TP predictions) should have higher confidence scores and be shifted to the right, indicating that the model is sure about its correct predictions, and the red histogram (FP predictions) should have lower confidence scores and be shifted to the left.
"""

markdown_class_ap_polar = (
    """## Average Precision by Class

A quick visual comparison of the model performance across all classes. Each axis in the chart represents a different class, and the distance to the center indicates the <abbr title="{}">Average Precision</abbr> (AP) for that class.

"""
    + clickable_label
)

markdown_class_ap_bar = (
    """## Average Precision by Class

A quick visual comparison of the model performance across all classes. Each bar in the chart represents a different class, and the height of the bar indicates the <abbr title="{}">Average Precision</abbr> (AP) for that class.

"""
    + clickable_label
)


markdown_class_outcome_counts_1 = """### Outcome Counts by Class

This chart breaks down all predictions into <abbr title="{}">True Positives</abbr> (TP), <abbr title="{}">False Positives</abbr> (FP), and <abbr title="{}">False Negatives</abbr> (FN) by classes. This helps to visually assess the type of errors the model often encounters for each class.

"""

markdown_normalization = """Normalization is used for better interclass comparison. If the normalization is on, the total outcome counts are divided by the number of ground truth instances of the corresponding class. This is useful, because on the chart, the sum of TP and FN bars will always result in 1.0, representing the full set of ground truth instances in the dataset for a class. This provides a clear visual understanding of how many instances the model correctly detected, how many it missed, and how many were false positives. For example, if a green bar (TP outcomes) reaches 1.0, this means the model has managed to predict all objects for the class without false negatives. Everything that is higher than 1.0 corresponds to False Positives, i.e, redundant predictions that the model should not predict. You can turn off the normalization, switching to absolute counts.

If normalization is off, the chart will display the total count of instances that correspond to outcome type (one of TP, FP or FN). This mode is identical to the main Outcome Counts graph on the top of the page. However, when normalization is off, you may encounter a class imbalance problem. Visually, bars that correspond to classes with many instances in the dataset will be much larger than others. This complicates the visual analysis.
"""

markdown_class_outcome_counts_2 = (
    """You can switch the plot view between normalized and absolute values.

_Bars in the chart are sorted by <abbr title="{}">F1-score</abbr> to keep a unified order of classes between different charts._

"""
    + clickable_label
)

empty = """### {}

> {}
"""

markdown_speedtest_intro = """## Inference Speed

This is a speed test benchmark for compared models. Models were tested with the following configurations:
"""

markdown_speedtest_overview_ms = """### Latency (Inference Time)
The table below shows the speed test results. For each test, the time taken to process one batch of images is shown. Results are averaged across **{}** iterations.
"""

markdown_speedtest_overview_fps = """### Frames per Second (FPS)
The table below shows the speed test results. For each test, the number of frames processed per second is shown. Results are averaged across **{}** iterations.
"""

markdown_batch_inference = """
This chart shows how the model's speed changes with different batch sizes . As the batch size increases, you can observe an increase in FPS (images per second).
"""



import random
import string
from abc import abstractmethod
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from json import JSONDecodeError
from os.path import dirname, join
from time import time
from typing import Any, Dict, List, Literal, NamedTuple

import requests

from supervisely import logger
from supervisely._utils import abs_url, is_development
from supervisely.api.api import Api, ApiField
from supervisely.api.file_api import FileInfo
from supervisely.io.fs import get_file_name_with_ext, silent_remove
from supervisely.io.json import dump_json_file
from supervisely.nn.experiments import ExperimentInfo


class TrainInfo(NamedTuple):
    """
    TrainInfo
    """

    app_name: str
    task_id: int
    artifacts_folder: str
    session_link: str
    task_type: str
    project_name: str
    checkpoints: List[FileInfo]
    config_path: str = None


class BaseTrainArtifacts:
    def __init__(self, team_id: int):
        """
        This is a base class and is not intended to be instantiated directly.
        Subclasses should implement the abstract methods.

        :param team_id: The team ID.
        :type team_id: int
        :raises Exception: If the class is instantiated directly.
        """
        if type(self) is BaseTrainArtifacts:
            raise Exception(
                "BaseTrainArtifacts is a base class and should not be instantiated directly"
            )

        self._api: Api = Api.from_env()
        self._team_id: int = team_id
        self._metadata_file_name: str = "train_info.json"

        self._app_name: str = None
        self._framework_name: str = None
        self._framework_folder: str = None
        self._weights_folder: str = None
        self._task_type: str = None
        self._weights_ext: str = None
        self._config_file: str = None
        self._pattern: str = None
        self._available_task_types: List[str] = []

    @property
    def team_id(self) -> int:
        """
        Get the team ID.

        :return: The team ID.
        :rtype: int
        """
        return self._team_id

    @property
    def metadata_file_name(self) -> str:
        """
        Metadata file name.

        :return: The metadata file name.
        :rtype: str
        """
        return self._metadata_file_name

    @property
    def app_name(self):
        """
        Train application name.

        :return: The train application name.
        :rtype: str
        """
        return self._app_name

    @property
    def framework_name(self):
        """
        Framework name.

        :return: The framework name.
        :rtype: str
        """
        return self._framework_name

    @property
    def framework_folder(self):
        """
        Path to framework folder in Supervisely Team Files.

        :return: The framework folder path.
        :rtype: str
        """
        return self._framework_folder

    @property
    def weights_folder(self):
        """
        Weights folder path relative to artifacts folder.

        :return: The weights folder path.
        :rtype: str
        """
        return self._weights_folder

    @property
    def task_type(self):
        """
        Framework computer vision task. None if can be multiple tasks.

        :return: The cv task.
        :rtype: Union[str, None]
        """
        return self._task_type

    @property
    def weights_ext(self):
        """
        Checkpoint weights extension.

        :return: The weights extension.
        :rtype: str
        """
        return self._weights_ext

    @property
    def config_file(self):
        """
        Name of the config file with extension.

        :return: The config file name.
        :rtype: str
        """
        return self._config_file

    @property
    def pattern(self):
        """
        Framework artifacts folder path pattern.

        :return: The artifacts folder path pattern.
        :rtype: re.Pattern
        """
        return self._pattern

    def is_valid_artifacts_path(self, path):
        """
        Check if the provided path is valid and follows specified session path pattern.

        :param path: Path to artifacts folder.
        :type path: str
        :return: True if the path to artifacts folder is valid, False otherwise.
        :rtype: bool
        """
        return self._pattern.match(path) is not None

    @abstractmethod
    def get_task_id(self, artifacts_folder: str) -> str:
        """
        Get the task ID of training session.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        :return: The task ID.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_project_name(self, artifacts_folder: str) -> str:
        """
        Get the training project name.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        :return: The training project name.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_task_type(self, artifacts_folder: str) -> str:
        """
        Get the cv task.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        :return: The cv task.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_weights_path(self, artifacts_folder: str) -> str:
        """
        Get path to weights folder.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        :return: The weights folder path.
        :rtype: str
        """
        pass

    @abstractmethod
    def get_config_path(self, artifacts_folder: str) -> str:
        """
        Get path to config file.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        :return: The config file path.
        :rtype: str
        """
        pass

    def sort_train_infos(
        self, train_infos: List[TrainInfo], sort: Literal["desc", "asc"] = "desc"
    ) -> List[TrainInfo]:
        """
        Sort artifacts folder by task id.

        :param train_infos: The list of training infos.
        :type train_infos: List[TrainInfo]
        :param sort: The sort order, either "desc" or "asc". Default is "desc", which means newer checkpoints will be first.
        :type sort: Literal["desc", "asc"]
        :return: The sorted list of checkpoints.
        :rtype: List[TrainInfo]
        """
        if sort == "desc":
            return sorted(train_infos, key=lambda x: int(x.task_id), reverse=True)
        elif sort == "asc":
            return sorted(train_infos, key=lambda x: int(x.task_id))

    def remove_metadata(self, artifacts_folder: str) -> None:
        """
        Remove the metadata file from the session folder.

        :param artifacts_folder: Path to artifacts folder.
        :type artifacts_folder: str
        """
        metadata_path = join(artifacts_folder, self._metadata_file_name)
        self._api.file.remove(self._team_id, metadata_path)
        logger.info(f"File '{metadata_path}' was removed")

    def remove_all_metadatas(self) -> None:
        """
        Remove the metadata files from the session folders.

        :param artifacts_folders: Path to artifacts folders.
        :type artifacts_folders: List[str]
        """
        count = 0
        file_paths = [file_info.path for file_info in self._get_file_infos()]
        for file_path in file_paths:
            if file_path.endswith(self._metadata_file_name):
                self._api.file.remove(self._team_id, file_path)
                logger.info(f"File '{file_path}' was removed")
                count += 1
        logger.info(f"Total files removed: '{count}'")

    def generate_metadata(
        self,
        app_name: str,
        task_id: str,
        artifacts_folder: str,
        weights_folder: str,
        weights_ext: str,
        project_name: str,
        task_type: str = None,
        config_path: str = None,
    ):
        """
        Generate the metadata for the given parameters.

        :param app_name: Name of the training application.
        :type app_name: str
        :param task_id: The session ID.
        :type task_id: str
        :param artifacts_folder: Path to session folder.
        :type artifacts_folder: str
        :param weights_folder: Path to weights location.
        :type weights_folder: str
        :param weights_ext: The weights extension.
        :type weights_ext: str
        :param project_name: Name of project used for training.
        :type project_name: str
        :param task_type: CV Task. Default is None.
        :type task_type: str, optional
        :param config_path: Path to config file. Default is None.
        :type config_path: str, optional
        :return: The generated metadata.
        :rtype: dict
        """

        def _get_checkpoint_file_infos(weights_folder) -> List[FileInfo]:
            return [
                file
                for file in self._api.file.list(
                    self._team_id,
                    weights_folder,
                    recursive=False,
                    return_type="fileinfo",
                )
                if file.name.endswith(weights_ext)
            ]

        def _upload_metadata(json_data: dict) -> None:
            random_string = "".join(random.choices(string.ascii_lowercase + string.digits, k=10))
            json_data_path = f"{random_string}.txt"
            dump_json_file(json_data, json_data_path)
            self._api.file.upload(
                self._team_id,
                json_data_path,
                f"{artifacts_folder}/{self._metadata_file_name}",
            )
            silent_remove(json_data_path)

        checkpoint_file_infos = _get_checkpoint_file_infos(weights_folder)
        if hasattr(self, "_legacy_weights_folder"):
            if self._legacy_weights_folder is not None:
                legacy_checkpoint_file_infos = _get_checkpoint_file_infos(
                    self._legacy_weights_folder
                )
                if len(legacy_checkpoint_file_infos) > 0:
                    checkpoint_file_infos.extend(legacy_checkpoint_file_infos)

        if len(checkpoint_file_infos) == 0:
            logger.info(f"No checkpoints found in '{weights_folder}'")
            return None

        logger.info(f"Generating '{self._metadata_file_name}' for '{artifacts_folder}'")
        if is_development():
            session_link = abs_url(f"/apps/sessions/{task_id}")
        else:
            session_link = f"/apps/sessions/{task_id}"

        train_json = {
            "app_name": app_name,
            "task_id": task_id,
            "artifacts_folder": artifacts_folder,
            "session_link": session_link,
            "task_type": task_type,
            "project_name": project_name,
            "checkpoints": checkpoint_file_infos,
        }
        if config_path is not None:
            train_json["config_path"] = config_path
        is_valid = self._validate_train_json(train_json)
        if is_valid:
            _upload_metadata(train_json)
            logger.info(f"Metadata for '{artifacts_folder}' was generated")
        else:
            logger.warn(f"Invalid metadata for '{artifacts_folder}'")
            train_json = None
        return train_json

    def _fetch_json_from_path(self, remote_path: str):
        try:
            response = self._api.post(
                "file-storage.download",
                {ApiField.TEAM_ID: self.team_id, ApiField.PATH: remote_path},
                stream=True,
            )
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            logger.debug(f"Failed to fetch train metadata from '{remote_path}': {e}")
            return None

        try:
            response_json = response.json()
        except JSONDecodeError as e:
            logger.debug(f"Failed to decode JSON from '{remote_path}': {e}")
            return None

        checkpoints = response_json.get("checkpoints", [])
        file_infos = [FileInfo(*checkpoint) for checkpoint in checkpoints]
        response_json["checkpoints"] = file_infos

        return response_json

    def _get_train_json(
        self,
        artifacts_folder: str,
        metadata_path: str,
        file_infos: List[FileInfo],
        file_paths: List[str],
    ) -> Dict[str, Any]:
        json_data = None
        if metadata_path not in file_paths:
            weights_folder = self.get_weights_path(artifacts_folder)
            task_type = self.get_task_type(artifacts_folder)
            task_id = self.get_task_id(artifacts_folder)
            project_name = self.get_project_name(artifacts_folder)
            config_path = self.get_config_path(artifacts_folder)
            json_data = self.generate_metadata(
                app_name=self._app_name,
                task_id=task_id,
                artifacts_folder=artifacts_folder,
                weights_folder=weights_folder,
                weights_ext=self._weights_ext,
                project_name=project_name,
                task_type=task_type,
                config_path=config_path,
            )
        else:
            start_find_time = time()
            for file_info in file_infos:
                if file_info.path == metadata_path:
                    json_data = self._fetch_json_from_path(file_info.path)
                    break
            end_find_time = time()
            logger.debug(
                f"Fetch metadata for {metadata_path}: '{format(end_find_time - start_find_time, '.6f')}' sec"
            )
            is_valid = self._validate_train_json(json_data)
            if not is_valid:
                logger.warn(f"Invalid metadata for '{artifacts_folder}'")
                json_data = None
        return json_data

    def _validate_sort(self, sort: Literal["desc", "asc"]):
        if sort not in ["desc", "asc"]:
            raise ValueError(f"Invalid sort value: {sort}")

    def _get_file_infos(self):
        return self._api.file.list(self._team_id, self._framework_folder, return_type="fileinfo")

    def _group_files_by_folder(self, file_infos: List[FileInfo]) -> Dict[str, List[FileInfo]]:
        folders = defaultdict(list)
        for file_info in file_infos:
            artifacts_folder = dirname(file_info.path)
            if self.is_valid_artifacts_path(artifacts_folder):
                folders[artifacts_folder].append(file_info)
        return folders

    def _validate_train_json(self, train_json) -> bool:
        if not isinstance(train_json, dict):
            # Invalid train_json format
            return False
        required_fields = [
            "app_name",
            "task_id",
            "artifacts_folder",
            "session_link",
            "task_type",
            "project_name",
            "checkpoints",
        ]
        if self.app_name == "Train MMDetection" or self.app_name == "Train MMDetection 3.0":
            required_fields.append("config_path")

        for field in required_fields:
            if field not in train_json:
                # Missing 'field' in train_json
                return False
            else:
                value = train_json.get(field)
                if value is None:
                    logger.debug(f"Field '{field}' is None")
                    # 'field' is None
                    return False
        return True

    def _create_train_infos(self, folders):
        train_infos = []

        def process_folder(artifacts_folder, file_infos):
            metadata_path = join(artifacts_folder, self._metadata_file_name)
            file_paths = [file_info.path for file_info in file_infos]
            train_json = self._get_train_json(
                artifacts_folder, metadata_path, file_infos, file_paths
            )
            if train_json:
                return TrainInfo(**train_json)
            return None

        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(process_folder, folder, infos): folder
                for folder, infos in folders.items()
            }
            for future in as_completed(futures):
                train_info = future.result()
                if train_info:
                    train_infos.append(train_info)

        return train_infos

    def get_list(self, sort: Literal["desc", "asc"] = "desc") -> List[TrainInfo]:
        """
        Return list of custom training infos

        :param sort: The sort order, either "desc" or "asc". Default is "desc".
        :type sort: Literal["desc", "asc"]
        :return: The list of custom training artifact infos.
        :rtype: List[TrainInfo]
        """
        self._validate_sort(sort)
        start_time = time()
        parsed_infos = self._get_file_infos()
        folders = self._group_files_by_folder(parsed_infos)

        with ThreadPoolExecutor() as executor:
            future = executor.submit(self._create_train_infos, folders)
            train_infos = future.result()

        end_time = time()
        train_infos = self.sort_train_infos(train_infos, sort)
        logger.debug(f"Listing time: '{format(end_time - start_time, '.6f')}' sec")
        return train_infos

    def get_list_experiment_info(self, sort: Literal["desc", "asc"] = "desc") -> List[TrainInfo]:
        def build_experiment_info_from_train_info(
            api: Api, train_info: TrainInfo
        ) -> ExperimentInfo:

            try:
                checkpoints = []
                for chk in train_info.checkpoints:
                    if self.weights_folder:
                        checkpoints.append(join(self.weights_folder, chk.name))
                    else:
                        checkpoints.append(chk.name)

                best_checkpoint = next(
                    (chk.name for chk in train_info.checkpoints if "best" in chk.name), None
                )
                if not best_checkpoint and checkpoints:
                    best_checkpoint = get_file_name_with_ext(checkpoints[-1])

                task_info = api.task.get_info_by_id(train_info.task_id)
                workspace_id = task_info["workspaceId"]

                project = api.project.get_info_by_name(workspace_id, train_info.project_name)
                project_id = project.id if project else None

                model_files = {}
                if train_info.config_path:
                    model_files["config"] = self.get_config_path(
                        train_info.artifacts_folder
                    ).replace(train_info.artifacts_folder, "")

                input_datetime = task_info["startedAt"]
                parsed_datetime = datetime.strptime(input_datetime, "%Y-%m-%dT%H:%M:%S.%fZ")
                date_time = parsed_datetime.strftime("%Y-%m-%d %H:%M:%S")

                experiment_info_data = {
                    "experiment_name": f"Unknown {self.framework_name} experiment",
                    "framework_name": self.framework_name,
                    "model_name": f"Unknown {self.framework_name} model",
                    "task_type": train_info.task_type,
                    "project_id": project_id,
                    "task_id": train_info.task_id,
                    "model_files": model_files,
                    "checkpoints": checkpoints,
                    "best_checkpoint": best_checkpoint,
                    "artifacts_dir": train_info.artifacts_folder,
                    "datetime": date_time,
                }

                experiment_info_fields = {
                    field.name
                    for field in ExperimentInfo.__dataclass_fields__.values()  # pylint: disable=no-member
                }
                for field in experiment_info_fields:
                    if field not in experiment_info_data:
                        experiment_info_data[field] = None

                return ExperimentInfo(**experiment_info_data)
            except Exception as e:
                logger.warning(f"Failed to build experiment info: {e}")
                return None

        train_infos = self.get_list(sort)

        # Sync version
        # Uncomment for debug
        # experiment_infos = [
        #     convert_traininfo_to_experimentinfo(api, framework_cls, train_info)
        #     for train_info in train_infos
        # ]
        # return experiment_infos

        # Async version
        with ThreadPoolExecutor() as executor:
            experiment_infos = list(
                executor.map(
                    lambda t: build_experiment_info_from_train_info(self._api, t),
                    train_infos,
                )
            )
        return [info for info in experiment_infos if info is not None]

    def get_available_task_types(self) -> List[str]:
        """
        Get available task types.

        :return: The list of available task types.
        :rtype: List[str]
        """
        return self._available_task_types



import os
import sys
import uuid
from copy import deepcopy
from typing import List

import cv2
import numpy as np


class HiddenCocoPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


from supervisely import (
    Annotation,
    Bitmap,
    GraphNodes,
    Label,
    Node,
    ObjClass,
    PointLocation,
    Polygon,
    ProjectMeta,
    Rectangle,
    Tag,
    TagMeta,
    TagValueType,
    logger,
)
from supervisely.convert.image.image_converter import ImageConverter
from supervisely.convert.image.image_helper import validate_image_bounds
from supervisely.geometry.graph import KeypointsTemplate
from supervisely.imaging.color import generate_rgb

conflict_classes = []


# COCO Convert funcs
def create_supervisely_annotation(
    item: ImageConverter.Item,
    meta: ProjectMeta,
    coco_categories: List[dict],
    renamed_classes: dict = None,
    renamed_tags: dict = None,
):
    labels = []
    imag_tags = []
    name_cat_id_map = coco_category_to_class_name(coco_categories)
    renamed_classes = {} if renamed_classes is None else renamed_classes
    renamed_tags = {} if renamed_tags is None else renamed_tags
    for coco_data in item.ann_data:
        caption = coco_data.get("caption")
        if caption is not None:
            tag_name = renamed_tags.get("caption", "caption")
            imag_tags.append(Tag(meta.get_tag_meta(tag_name), caption))
        category_id = coco_data.get("category_id")
        if category_id is None:
            continue
        obj_class_name = name_cat_id_map.get(category_id)
        if obj_class_name is None:
            logger.warning(f"Category with id {category_id} not found in categories list")
            continue
        renamed_class_name = renamed_classes.get(obj_class_name, obj_class_name)
        key = None

        segm = coco_data.get("segmentation")
        keypoints = coco_data.get("keypoints")
        bbox = coco_data.get("bbox")

        if len([f for f in (segm, keypoints, bbox) if f]) > 1:
            # create a binding key if more than one of the following fields are present
            key = uuid.uuid4().hex

        curr_labels = []
        if segm is not None and len(segm) > 0:
            obj_class_polygon = meta.get_obj_class(renamed_class_name)
            if obj_class_polygon.geometry_type != Polygon:
                if obj_class_name not in conflict_classes:
                    geometry_name = obj_class_polygon.geometry_type.geometry_name().capitalize()
                    conflict_classes.append(obj_class_name)
                    logger.warning(
                        "Conflict in class geometry type: "
                        f"object class '{obj_class_name}' (category ID: {category_id}) "
                        f"has type '{geometry_name}', but expected type is 'Polygon'."
                    )
                continue
            if type(segm) is dict:
                polygons = convert_rle_mask_to_polygon(coco_data)
                for polygon in polygons:
                    curr_labels.append(Label(polygon, obj_class_polygon, binding_key=key))
            elif type(segm) is list and coco_data["segmentation"]:
                polygons = convert_polygon_vertices(coco_data, item.shape)
                for polygon in polygons:
                    curr_labels.append(Label(polygon, obj_class_polygon, binding_key=key))

        if keypoints is not None:
            obj_class_keypoints = meta.get_obj_class(renamed_class_name)
            keypoints = list(get_coords(keypoints))
            coco_categorie, keypoint_names = None, None
            for cat in coco_categories:
                if cat["id"] == category_id and cat["supercategory"] == obj_class_name:
                    coco_categorie = cat
                    break
            if coco_categorie is not None:
                keypoint_names = coco_categorie.get("keypoints")
            if keypoint_names is not None:
                nodes = []
                for coords, keypoint_name in zip(keypoints, keypoint_names):
                    col, row, visibility = coords
                    if visibility in [0, 1]:
                        continue  # skip invisible keypoints

                    node = Node(label=keypoint_name, row=row, col=col)  # , disabled=v)
                    nodes.append(node)
                if len(nodes) != 0:
                    label = Label(GraphNodes(nodes), obj_class_keypoints, binding_key=key)
                    curr_labels.append(label)
        labels.extend(curr_labels)

        if bbox is not None and len(bbox) == 4:
            if not obj_class_name.endswith("bbox"):
                obj_class_name = add_tail(obj_class_name, "bbox")
            renamed_class_name = renamed_classes.get(obj_class_name, obj_class_name)
            obj_class_rectangle = meta.get_obj_class(renamed_class_name)
            if obj_class_rectangle.geometry_type != Rectangle:
                if obj_class_name not in conflict_classes:
                    geometry_name = obj_class_rectangle.geometry_type.geometry_name().capitalize()
                    conflict_classes.append(obj_class_name)
                    logger.warning(
                        "Conflict in class geometry type: "
                        f"object class '{obj_class_name}' (category ID: {category_id}) "
                        f"has type '{geometry_name}', but expected type is 'Rectangle'."
                    )
                continue
            x, y, w, h = bbox
            geometry = Rectangle(y, x, y + h, x + w)
            labels.append(Label(geometry, obj_class_rectangle, binding_key=key))
    labels = validate_image_bounds(labels, Rectangle.from_size(item.shape))
    return Annotation(item.shape, labels=labels, img_tags=imag_tags)


def convert_rle_mask_to_polygon(coco_ann):
    import pycocotools.mask as mask_util  # pylint: disable=import-error

    if type(coco_ann["segmentation"]["counts"]) is str:
        coco_ann["segmentation"]["counts"] = bytes(
            coco_ann["segmentation"]["counts"], encoding="utf-8"
        )
        mask = mask_util.decode(coco_ann["segmentation"])
    else:
        rle_obj = mask_util.frPyObjects(
            coco_ann["segmentation"],
            coco_ann["segmentation"]["size"][0],
            coco_ann["segmentation"]["size"][1],
        )
        mask = mask_util.decode(rle_obj)
    mask = np.array(mask, dtype=bool)
    if not np.any(mask):
        return []
    return Bitmap(mask).to_contours()


def convert_polygon_vertices(coco_ann, image_size):
    polygons = coco_ann["segmentation"]
    if all(type(coord) is float for coord in polygons):
        polygons = [polygons]
    if any(type(coord) is str for polygon in polygons for coord in polygon):
        return []
    exteriors = []
    for polygon in polygons:
        polygon = [polygon[i * 2 : (i + 1) * 2] for i in range((len(polygon) + 2 - 1) // 2)]
        exterior_points = [(width, height) for width, height in polygon]
        if len(exterior_points) == 0:
            continue
        exteriors.append(exterior_points)
    interiors = {idx: [] for idx in range(len(exteriors))}
    id2del = []
    for idx, exterior in enumerate(exteriors):
        temp_img = np.zeros(image_size + (3,), dtype=np.uint8)
        geom = Polygon([PointLocation(y, x) for x, y in exterior])
        geom.draw_contour(temp_img, color=[255, 255, 255])
        im = cv2.cvtColor(temp_img, cv2.COLOR_RGB2GRAY)
        contours, _ = cv2.findContours(im, cv2.RETR_LIST, cv2.CHAIN_APPROX_SIMPLE)
        if len(contours) == 0:
            continue
        for idy, exterior2 in enumerate(exteriors):
            if idx == idy or idy in id2del:
                continue
            points_inside = [
                cv2.pointPolygonTest(contours[0], (x, y), False) > 0 for x, y in exterior2
            ]
            if all(points_inside):
                interiors[idx].append(deepcopy(exteriors[idy]))
                id2del.append(idy)
    for j in sorted(id2del, reverse=True):
        del exteriors[j]
    figures = []
    for exterior, interior in zip(exteriors, interiors.values()):
        exterior = [PointLocation(y, x) for x, y in exterior]
        interior = [[PointLocation(y, x) for x, y in points] for points in interior]
        figures.append(Polygon(exterior, interior))
    return figures


def generate_meta_from_annotation(coco, meta: ProjectMeta = None):
    from pycocotools.coco import COCO  # pylint: disable=import-error

    coco: COCO

    colors = []
    ann_types = get_ann_types(coco)
    if len(coco.cats) == 0:
        categories = []
    else:
        categories = coco.loadCats(ids=coco.getCatIds())
    if meta is None:
        meta = ProjectMeta()
    for category in categories:
        if category["name"] in [obj_class.name for obj_class in meta.obj_classes]:
            continue
        obj_classes = []
        new_color = generate_rgb(colors)
        colors.append(new_color)
        if ann_types is not None:  # add skeleton support
            if "keypoints" in ann_types and "keypoints" in category:
                geometry_config = None
                cat_num_kp = None
                cat_labels = category.get("keypoints", None)
                cat_edges = category.get("skeleton", None)
                if cat_labels is None:
                    cat_num_kp = []
                    for img_id, img_info in coco.imgs.items():
                        ann_p = coco.anns[img_id]
                        for ann in ann_p:
                            if ann["category_id"] == category["id"]:
                                cat_num_kp.append(ann["num_keypoints"])
                    cat_num_kp = max(cat_num_kp)

                geometry_config = create_custom_geometry_config(
                    num_keypoints=cat_num_kp, cat_labels=cat_labels, cat_edges=cat_edges
                )
                obj_class = ObjClass(
                    name=category["name"],
                    geometry_type=GraphNodes,
                    color=new_color,
                    geometry_config=geometry_config,
                )
                obj_classes.append(obj_class)
            elif "segmentation" in ann_types:
                obj_classes.append(ObjClass(category["name"], Polygon, new_color))
            if "bbox" in ann_types:
                obj_classes.append(
                    ObjClass(add_tail(category["name"], "bbox"), Rectangle, new_color)
                )
        for obj_class in obj_classes:
            existing_classes = [obj_class.name for obj_class in meta.obj_classes]
            if obj_class.name not in existing_classes:
                meta = meta.add_obj_class(obj_class)
    if ann_types is not None and "caption" in ann_types:
        tag_meta = TagMeta("caption", TagValueType.ANY_STRING)
        if meta.get_tag_meta(tag_meta.name) is None:
            meta = meta.add_tag_meta(tag_meta)
    return meta


def get_ann_types(coco) -> List[str]:
    from pycocotools.coco import COCO  # pylint: disable=import-error

    coco: COCO

    ann_types = []
    annotation_ids = coco.getAnnIds()
    if any("bbox" in coco.anns[ann_id] for ann_id in annotation_ids):
        ann_types.append("bbox")
    if any("segmentation" in coco.anns[ann_id] for ann_id in annotation_ids):
        ann_types.append("segmentation")
    if any("caption" in coco.anns[ann_id] for ann_id in annotation_ids):
        ann_types.append("caption")
    if any("keypoints" in coco.anns[ann_id] for ann_id in annotation_ids):
        ann_types.append("keypoints")
    return ann_types


def add_tail(body: str, tail: str):
    if " " in body:
        return f"{body} {tail}"
    return f"{body}_{tail}"


def coco_category_to_class_name(coco_categories):
    return {category["id"]: category["name"] for category in coco_categories}


def get_coords(keypoints):
    for i in range(0, len(keypoints), 3):
        yield keypoints[i : i + 3]


def create_custom_geometry_config(num_keypoints=None, cat_labels=None, cat_edges=None):
    template = KeypointsTemplate()

    if cat_labels is None:
        if num_keypoints is None:
            raise ValueError(
                "Number of keypoints can not be specified, please check your annotation (categories: num_keypoints)"
            )
        for p in list(range(num_keypoints)):
            template.add_point(label=str(p), row=0, col=0)
    else:
        for label in cat_labels:
            template.add_point(label=label, row=0, col=0)

    if cat_edges is not None and cat_labels is not None:
        for edge in cat_edges:
            template.add_edge(src=cat_labels[edge[0] - 1], dst=cat_labels[edge[1] - 1])
    else:
        logger.warn("Edges can not be mapped without skeleton, please check your annotation")
    return template



from typing import Any, Dict, Optional

import numpy as np

METRIC_NAMES = {
    "mPixel": "mPixel accuracy",
    "mF1": "mF1-score",
    "mPrecision": "mPrecision",
    "mRecall": "mRecall",
    "mIoU": "mIoU",
    "mBoundaryIoU": "mBoundaryIoU",
    "calibration_score": "Calibration Score",
}


class MetricProvider:
    def __init__(self, eval_data: Dict[str, Any]):
        # self.params = params
        self.metric_names = METRIC_NAMES

        # eval_data
        self.eval_data = eval_data["result"]
        self.bg_cls_name = eval_data["bg_cls_name"]
        self.per_image_metrics = eval_data["per_image_metrics"]
        self.cmat_cell_img_names = eval_data["cell_img_names"]
        self.class_names = self.eval_data.index.tolist()

        self.num_classes = len(self.class_names)

        # base metrics
        overall_TP = self.eval_data["TP"][: self.num_classes].sum()
        overall_FN = self.eval_data["FN"][: self.num_classes].sum()
        self.pixel_accuracy = overall_TP / (overall_TP + overall_FN)
        self.overall_TP = overall_TP
        self.overall_FN = overall_FN
        self.precision = round(self.eval_data.loc["mean", "precision"] * 100, 1)
        self.recall = round(self.eval_data.loc["mean", "recall"] * 100, 1)
        self.f1_score = round(self.eval_data.loc["mean", "F1_score"] * 100, 1)
        self.iou = round(self.eval_data.loc["mean", "IoU"] * 100, 1)
        self.boundary_iou = round(self.eval_data.loc["mean", "boundary_IoU"] * 100, 1)

        # error metrics
        # labels = ["mIoU", "mBoundaryEoU", "mExtentEoU", "mSegmentEoU"]
        self.boundary_eou = round(self.eval_data.loc["mean", "E_boundary_oU"] * 100, 1)
        self.extent_eou = round(self.eval_data.loc["mean", "E_extent_oU"] * 100, 1)
        self.segment_eou = round(self.eval_data.loc["mean", "E_segment_oU"] * 100, 1)

        # renormalized error metrics
        # labels = ["boundary", "extent", "segment"]
        self.boundary_renormed_eou = round(
            self.eval_data.loc["mean", "E_boundary_oU_renormed"] * 100, 1
        )
        self.extent_renormed_eou = round(
            self.eval_data.loc["mean", "E_extent_oU_renormed"] * 100, 1
        )
        self.segment_renormed_eou = round(
            self.eval_data.loc["mean", "E_segment_oU_renormed"] * 100, 1
        )

        # classwise error data
        self.classwise_segm_error_data = self.get_classwise_error_data()

        # confusion matrix
        self.confusion_matrix = self.get_confusion_matrix(eval_data["confusion_matrix"].copy())

        # frequently confused classes
        self.frequently_confused = self.get_frequently_confused(
            eval_data["confusion_matrix"].copy(), n_pairs=20
        )

    def json_metrics(self):
        return {
            "mIoU": self.eval_data.loc["mean"]["IoU"] * 100,
            "mE_boundary_oU": self.eval_data.loc["mean"]["E_boundary_oU"] * 100,
            "mFP_boundary_oU": self.eval_data.loc["mean"]["FP_boundary_oU"] * 100,
            "mFN_boundary_oU": self.eval_data.loc["mean"]["FN_boundary_oU"] * 100,
            "mE_boundary_oU_renormed": self.eval_data.loc["mean"]["E_boundary_oU_renormed"] * 100,
            "mE_extent_oU": self.eval_data.loc["mean"]["E_extent_oU"] * 100,
            "mFP_extent_oU": self.eval_data.loc["mean"]["FP_extent_oU"] * 100,
            "mFN_extent_oU": self.eval_data.loc["mean"]["FN_extent_oU"] * 100,
            "mE_extent_oU_renormed": self.eval_data.loc["mean"]["E_extent_oU_renormed"] * 100,
            "mE_segment_oU": self.eval_data.loc["mean"]["E_segment_oU"] * 100,
            "mFP_segment_oU": self.eval_data.loc["mean"]["FP_segment_oU"] * 100,
            "mFN_segment_oU": self.eval_data.loc["mean"]["FN_segment_oU"] * 100,
            "mE_segment_oU_renormed": self.eval_data.loc["mean"]["E_segment_oU_renormed"] * 100,
            "mPrecision": self.eval_data.loc["mean"]["precision"] * 100,
            "mRecall": self.eval_data.loc["mean"]["recall"] * 100,
            "mF1_score": self.eval_data.loc["mean"]["F1_score"] * 100,
            "PixelAcc": self.pixel_accuracy * 100,
            "mBoundaryIoU": self.eval_data.loc["mean"]["boundary_IoU"] * 100,
        }

    def metric_table(self):
        names_map = {
            "img_names": "Image name",
            "pixel_acc": "Pixel accuracy",
            "precision": "Precision",
            "recall": "Recall",
            "f1_score": "F1 score",
            "iou": "IoU",
            "boundary_iou": "Boundary IoU",
            "boundary_eou": "Boundary EoU",
            "extent_eou": "Extent EoU",
            "segment_eou": "Segment EoU",
            "boundary_eou_renormed": "Boundary EoU renormed",
            "extent_eou_renormed": "Extent EoU renormed",
            "segment_eou_renormed": "Segment EoU renormed",
        }
        prediction_table = self.per_image_metrics.rename(columns=names_map)
        return prediction_table

    def get_confusion_matrix(self, confusion_matrix: np.ndarray):
        class_names = self.eval_data.index.tolist()
        confusion_matrix = confusion_matrix[::-1]
        return confusion_matrix, class_names

    def get_frequently_confused(self, confusion_matrix: np.ndarray, n_pairs: Optional[int] = None):

        non_diagonal_ids = {}
        for i, idx in enumerate(np.ndindex(confusion_matrix.shape)):
            if idx[0] != idx[1]:
                non_diagonal_ids[i] = idx

        indexes_1d = np.argsort(confusion_matrix, axis=None)
        indexes_2d = [non_diagonal_ids[idx] for idx in indexes_1d if idx in non_diagonal_ids]
        if n_pairs is not None:
            indexes_2d = indexes_2d[:n_pairs]
        indexes_2d = np.asarray(indexes_2d[::-1])

        rows = indexes_2d[:, 0]
        cols = indexes_2d[:, 1]
        probs = confusion_matrix[rows, cols]
        return probs, indexes_2d

    def key_metrics(self):
        return {
            "mPixel accuracy": self.pixel_accuracy,
            "mPrecision": self.precision,
            "mRecall": self.recall,
            "mF1-score": self.f1_score,
            "mIoU": self.iou,
            "mBoundaryIoU": self.boundary_iou,
            "mPixel accuracy": self.pixel_accuracy,
        }

    def get_classwise_error_data(self):
        self.eval_data.drop(["mean"], inplace=True)
        bar_data = self.eval_data.copy()
        bar_data = bar_data[["IoU", "E_extent_oU", "E_boundary_oU", "E_segment_oU"]]
        bar_data.sort_values(by="IoU", ascending=False, inplace=True)
        labels = list(bar_data.index)
        return bar_data, labels



class TaskType:
    OBJECT_DETECTION = "object detection"
    INSTANCE_SEGMENTATION = "instance segmentation"
    SEMANTIC_SEGMENTATION = "semantic segmentation"
    POSE_ESTIMATION = "pose estimation"
    SALIENT_OBJECT_SEGMENTATION = "salient object segmentation"
    PROMPT_BASED_OBJECT_DETECTION = "prompt-based object detection"
    INTERACTIVE_SEGMENTATION = "interactive segmentation"
    PROMPTABLE_SEGMENTATION = "promptable segmentation"
    TRACKING = "tracking"
    OBJECT_DETECTION_3D = "object detection 3d"



import argparse
import asyncio
import os
import signal
import sys
import time

import supervisely as sly

LOG_LEVEL = "INFO"
# LOG_LEVEL = "DEBUG"
PROJECT_ID = 42112  #  41862
DATSET_ID = 98429
home_dir = os.path.expanduser("~")
common_path = os.path.join(home_dir, "test_project_download/")
save_path = os.path.join(common_path, "sync/")
save_path_async = os.path.join(common_path, "async/")
sly.fs.ensure_base_path(common_path)
api = sly.Api.from_env()

sly.fs.clean_dir(common_path)


def main_dpa(project_id: int, semaphore_size: int):
    sema = asyncio.Semaphore(semaphore_size)
    start = time.time()
    loop = sly.utils.get_or_create_event_loop()
    loop.run_until_complete(
        sly.VideoProject.download_async(api, project_id, save_path_async, semaphore=sema)
    )
    finish = time.time() - start
    print(f"Time taken for async method: {finish}")
    print(f"Project downloaded to {save_path_async}")


def main_dps(project_id: int):
    start = time.time()
    sly.VideoProject.download(api, project_id, save_path)
    finish = time.time() - start
    print(f"Time taken for old method: {finish}")


def compare_downloads(project_id: int, semaphore_size: int):
    main_dps(project_id)
    sly.fs.clean_dir(save_path)
    main_dpa(project_id, semaphore_size)
    sly.fs.clean_dir(save_path_async)


if __name__ == "__main__":
    # parser = argparse.ArgumentParser(description="Download and save project as files (async)")
    # parser.add_argument(
    #     "--server", type=str, default=os.environ["SERVER_ADDRESS"], help="Server address"
    # )
    # parser.add_argument("--token", type=str, default=os.environ["API_TOKEN"], help="API token")
    # parser.add_argument("--id", type=int, default=PROJECT_ID, help="ID of the project to download")
    # parser.add_argument(
    #     "--semaphore", type=int, default=50, help="Semaphore size for async download"
    # )
    # args = parser.parse_args()
    try:
        # ann_db()
        # api = sly.Api(args.server, args.token)
        # api.logger.setLevel(LOG_LEVEL)
        # main_dpa(PROJECT_ID, 200)  # to download and save project as files (async)
        # main_dps(args.project_id)  # to download and save project as files (sync)
        compare_downloads(
            PROJECT_ID, 200
        )  # to compare the time taken for downloading and saving project as files (sync vs async)
    except KeyboardInterrupt:
        sly.logger.info("Stopped by user")



import os
from collections import defaultdict
from typing import Dict, Iterable, List, Optional, Union

import cv2
import numpy as np
import pandas as pd

from supervisely.nn.benchmark.utils.semantic_segmentation.utils import (
    dilate_mask,
    get_contiguous_segments,
    get_exterior_boundary,
    get_interior_boundary,
    get_single_contiguous_segment,
    one_hot,
    single_one_hot,
)
from supervisely.sly_logger import logger
from supervisely.task.progress import tqdm_sly

ERROR_CODES = {
    "ignore": -1,
    "unassigned": 0,
    "TP": 1,
    "TN": 2,
    "FP_boundary": 3,
    "FN_boundary": 4,
    "FP_extent": 5,
    "FN_extent": 6,
    "FP_segment": 7,
    "FN_segment": 8,
}


ERROR_PALETTE = {
    -1: (100, 100, 100),
    0: (150, 150, 150),
    1: (255, 255, 255),
    2: (0, 0, 0),
    3: (255, 200, 150),
    4: (150, 200, 255),
    5: (255, 100, 150),
    6: (150, 100, 255),
    7: (255, 0, 0),
    8: (0, 0, 255),
}


class Evaluator:
    def __init__(
        self,
        class_names: List[str],
        boundary_width: Union[float, int] = 0.01,
        boundary_iou_d: float = 0.02,
        boundary_implementation: str = "exact",
        result_dir: str = "output",
        progress: Optional[tqdm_sly] = None,
    ):
        """The main class for running our error analysis.
        :param class_names: List of strings providing names for class ids 0,...,C.
        :param boundary_width: The parameter d in the paper, either as a float in [0,1] (relative to diagonal)
            or as an integer > 1 (absolute number of pixels).
        :param boundary_implementation: Choose "exact" for the euclidean pixel distance.
            The Boundary IoU paper uses the L1 distance ("fast").
        """
        global torch, np, GPU, numpy
        import torch  # pylint: disable=import-error

        numpy = np
        GPU = False

        self.progress = progress or tqdm_sly
        self.class_names = class_names
        self.num_classes = len(self.class_names)

        self.boundary_width = boundary_width
        if 0 < self.boundary_width < 1:
            self.use_relative_boundary_width = True
        elif self.boundary_width % 1 != 0 or self.boundary_width < 0:
            raise ValueError("boundary_width should be an integer or a float in (0,1)!")
        else:
            self.use_relative_boundary_width = False

        self.boundary_implementation = boundary_implementation
        self.boundary_iou_d = boundary_iou_d

        self.confusion_matrix = np.zeros(
            (self.num_classes, self.num_classes),
        )
        self.cell_img_names = defaultdict(list)
        self.result_dir = result_dir

        self.image_metrics = {
            "img_names": [],
            "pixel_acc": [],
            "precision": [],
            "recall": [],
            "f1_score": [],
            "iou": [],
            "boundary_iou": [],
            "boundary_eou": [],
            "extent_eou": [],
            "segment_eou": [],
            "boundary_eou_renormed": [],
            "extent_eou_renormed": [],
            "segment_eou_renormed": [],
        }
        self.img_names = []
        self.per_image_metrics = pd.DataFrame()

    def extract_masks(self, seg, cl, n_cl):
        if GPU:
            seg = np.asarray(seg)
        h, w = seg.shape
        masks = np.zeros((n_cl, h, w))

        for i, c in enumerate(cl):
            masks[i, :, :] = seg == c

        return masks

    def extract_masks_gen(self, seg, cl):
        if GPU:
            seg = np.asarray(seg)
        h, w = seg.shape
        for c in cl:
            mask = np.zeros((h, w))
            mask[seg == c] = 1
            yield mask

    def calc_confusion_matrix(self, pred, gt, cmat, img_name):
        assert pred.shape == gt.shape

        cl = np.arange(1, self.num_classes + 1)
        for ig, gm in enumerate(self.extract_masks_gen(gt, cl)):
            if np.sum(gm) == 0:
                continue
            for ip, pm in enumerate(self.extract_masks_gen(pred, cl)):
                if np.sum(pm) == 0:
                    continue
                cmat[ig, ip] += np.sum(np.logical_and(pm, gm))
                self.cell_img_names[str(ig) + "_" + str(ip)].append(img_name)

        return cmat

    def evaluate(self, loader: Iterable):
        """This runs the analysis for a whole dataset.
        :param loader: Iterable providing pairs of (pred, gt).
        :returns: beyond_iou.Result.
        """
        self.results = {
            "unassigned": np.zeros(self.num_classes, dtype=np.int64),
            "ignore": np.zeros(self.num_classes, dtype=np.int64),
            "TP": np.zeros(self.num_classes, dtype=np.int64),
            "TN": np.zeros(self.num_classes, dtype=np.int64),
            "FP_boundary": np.zeros(self.num_classes, dtype=np.int64),
            "FN_boundary": np.zeros(self.num_classes, dtype=np.int64),
            "FP_extent": np.zeros(self.num_classes, dtype=np.int64),
            "FN_extent": np.zeros(self.num_classes, dtype=np.int64),
            "FP_segment": np.zeros(self.num_classes, dtype=np.int64),
            "FN_segment": np.zeros(self.num_classes, dtype=np.int64),
        }
        self.boundary_iou_intersection_counts = np.zeros(self.num_classes, dtype=np.int64)
        self.boundary_iou_union_counts = np.zeros(self.num_classes, dtype=np.int64)

        with self.progress(message="Calculating metrics...", total=len(loader)) as pbar:
            for pred, gt, img_name in loader:
                sample_results = self.evaluate_sample(pred, gt, img_name)
                self.update_results(sample_results, img_name)
                self.confusion_matrix = self.calc_confusion_matrix(
                    pred,
                    gt,
                    self.confusion_matrix,
                    img_name,
                )
                pbar.update(1)
        if GPU:
            for key, value in self.results.items():
                self.results[key] = value.get()  # pylint: disable=no-member
            self.boundary_iou_intersection_counts = self.boundary_iou_intersection_counts.get()
            self.boundary_iou_union_counts = self.boundary_iou_union_counts.get()

        result = self.calculate_error_metrics()
        normalized_confusion_matrix = self.confusion_matrix / self.confusion_matrix.sum(
            axis=1, keepdims=True
        )
        normalized_confusion_matrix[np.isnan(normalized_confusion_matrix)] = 0
        normalized_confusion_matrix = np.round(normalized_confusion_matrix, 3)
        self.per_image_metrics = pd.DataFrame(self.image_metrics, index=self.img_names)
        return {
            "result": result,
            "confusion_matrix": normalized_confusion_matrix,
            "per_image_metrics": self.per_image_metrics,
            "cell_img_names": self.cell_img_names,
        }

    def evaluate_sample(self, pred, gt, img_name):
        """Runs the analysis for a single sample.
        :param pred: Predicted segmentation as a numpy array of shape (H,W).
        :param gt: Ground-truth segmentation as a numpy array of shape (H,W).
        :returns: Dictionary holding results for this sample.
        """
        if pred.shape != gt.shape:
            raise RuntimeError(
                f"Shapes of prediction and annotation do not match! Pred: {pred.shape}, GT: {gt.shape}"
            )
        H, W = pred.shape
        results = np.full(
            shape=(self.num_classes, H, W),
            fill_value=ERROR_CODES["unassigned"],
            dtype=np.int8,
        )

        boundary_intersection_counts = np.zeros(self.num_classes, dtype=np.int64)
        boundary_union_counts = np.zeros(self.num_classes, dtype=np.int64)
        for c in range(self.num_classes):
            gt_mask = gt == c + 1
            pred_mask = pred == c + 1
            if not gt_mask.any() and not pred_mask.any():
                results[c] = ERROR_CODES["TN"]
                continue
            tp_mask = np.logical_and(gt_mask, pred_mask)
            tn_mask = ~np.logical_or(gt_mask, pred_mask)
            fp_mask = np.logical_and(pred_mask, ~gt_mask)
            fn_mask = np.logical_and(~pred_mask, gt_mask)
            results[c][tp_mask] = ERROR_CODES["TP"]
            results[c][tn_mask] = ERROR_CODES["TN"]
            # results[c][fp_mask] = ERROR_CODES["FP"]
            # results[c][fn_mask] = ERROR_CODES["FN"]

            # BOUNDARY
            results[c] = self.get_single_boundary_errors(
                class_results=results[c],
                tp_mask=tp_mask,
                tn_mask=tn_mask,
                fp_mask=fp_mask,
                fn_mask=fn_mask,
            )

            # EXTENT / SEGMENT
            results[c] = self.get_single_extent_segment_errors(
                class_results=results[c],
                pred_mask=pred_mask,
                gt_mask=gt_mask,
            )

            assert not (results[c] == ERROR_CODES["unassigned"]).any()

            # Boundary IoU
            ignore_inds = None
            (
                boundary_intersection_counts_active,
                boundary_union_counts_active,
            ) = self.evaluate_single_sample_boundary_iou(
                class_results=results[c],
                pred_mask=pred_mask,
                gt_mask=gt_mask,
                ignore_inds=ignore_inds,
            )

            boundary_intersection_counts[c] += boundary_intersection_counts_active
            boundary_union_counts[c] += boundary_union_counts_active

        return dict(
            main_results=results,
            boundary_iou_results=(boundary_intersection_counts, boundary_union_counts),
        )

    def update_results(self, sample_results, img_name):
        # main results
        image_stats = {}
        for error_name, error_code in ERROR_CODES.items():
            error_values = (sample_results["main_results"] == error_code).sum(axis=(1, 2))
            self.results[error_name] += error_values
            image_stats[error_name] = error_values

        # boundary IoU
        boundary_intersection_counts, boundary_union_counts = sample_results["boundary_iou_results"]
        self.boundary_iou_intersection_counts += boundary_intersection_counts
        self.boundary_iou_union_counts += boundary_union_counts
        image_stats["boundary_iou_intersection_counts"] = boundary_intersection_counts
        image_stats["boundary_iou_union_counts"] = boundary_union_counts
        self.calculate_per_image_metrics(image_stats, img_name)

    def calculate_per_image_metrics(self, image_stats, img_name):
        fp = image_stats["FP_boundary"] + image_stats["FP_extent"] + image_stats["FP_segment"]
        fn = image_stats["FN_boundary"] + image_stats["FN_extent"] + image_stats["FN_segment"]
        tp = image_stats["TP"]
        tn = image_stats["TN"]

        e_boundary = image_stats["FP_boundary"] + image_stats["FN_boundary"]
        e_extent = image_stats["FP_extent"] + image_stats["FN_extent"]
        e_segment = image_stats["FP_segment"] + image_stats["FN_segment"]

        union = tp + fp + fn
        iou = tp / union

        overall_tp = tp[: self.num_classes].sum()
        overall_fn = fn[: self.num_classes].sum()
        pixel_acc = overall_tp / (overall_tp + overall_fn)

        precision = tp / (tp + fp)
        recall = tp / (tp + fn)
        f1_score = 2 / (1.0 / precision + 1.0 / recall)

        fp_boundary_ou = image_stats["FP_boundary"] / union
        fn_boundary_ou = image_stats["FN_boundary"] / union
        e_boundary_ou = e_boundary / union

        fp_extent_ou = image_stats["FP_extent"] / union
        fn_extent_ou = image_stats["FN_extent"] / union
        e_extent_ou = e_extent / union

        fp_segment_ou = image_stats["FP_segment"] / union
        fn_segment_ou = image_stats["FN_segment"] / union
        e_segment_ou = e_segment / union

        e_boundary_ou_renormed = e_boundary / (tp + e_boundary)
        e_extent_ou_renormed = e_extent / (tp + e_boundary + e_extent)
        e_segment_ou_renormed = e_segment_ou

        with numpy.errstate(invalid="ignore"):
            boundary_iou = (
                image_stats["boundary_iou_intersection_counts"]
                / image_stats["boundary_iou_union_counts"]
            )

        def postprocess_values(values):
            values = values[~numpy.isnan(values)]
            value = round(float(np.mean(values)), 2)
            return value

        self.image_metrics["pixel_acc"].append(round(float(pixel_acc), 2))
        self.image_metrics["precision"].append(postprocess_values(precision))
        self.image_metrics["recall"].append(postprocess_values(recall))
        self.image_metrics["f1_score"].append(postprocess_values(f1_score))
        self.image_metrics["iou"].append(postprocess_values(iou))
        self.image_metrics["boundary_iou"].append(postprocess_values(boundary_iou))
        self.image_metrics["boundary_eou"].append(postprocess_values(e_boundary_ou))
        self.image_metrics["extent_eou"].append(postprocess_values(e_extent_ou))
        self.image_metrics["segment_eou"].append(postprocess_values(e_segment_ou))
        self.image_metrics["boundary_eou_renormed"].append(
            postprocess_values(e_boundary_ou_renormed)
        )
        self.image_metrics["extent_eou_renormed"].append(postprocess_values(e_extent_ou_renormed))
        self.image_metrics["segment_eou_renormed"].append(postprocess_values(e_segment_ou_renormed))
        self.image_metrics["img_names"].append(img_name)

        self.img_names.append(img_name)

    def get_boundary_errors(self, results, tp_mask, tn_mask, fp_mask, fn_mask):
        H, W = tp_mask.shape[-2:]
        if self.use_relative_boundary_width:
            img_diag = np.sqrt(H**2 + W**2)
            if GPU:
                img_diag = img_diag.get()
                tp_mask = tp_mask.get()
                tn_mask = tn_mask.get()

            boundary_width = int(round(self.boundary_width * img_diag))
        else:
            boundary_width = self.boundary_width

        tp_ext_boundary = get_exterior_boundary(
            tp_mask, width=boundary_width, implementation=self.boundary_implementation
        )
        tn_ext_boundary = get_exterior_boundary(
            tn_mask, width=boundary_width, implementation=self.boundary_implementation
        )

        if GPU:
            tp_ext_boundary, tn_ext_boundary = np.asarray(tp_ext_boundary), np.asarray(
                tn_ext_boundary
            )

        boundary_intersection = np.logical_and(tp_ext_boundary, tn_ext_boundary)
        fp_boundary_mask_naive = np.logical_and(fp_mask, boundary_intersection)
        fn_boundary_mask_naive = np.logical_and(fn_mask, boundary_intersection)

        if GPU:
            fp_boundary_mask_naive, fn_boundary_mask_naive = (
                fp_boundary_mask_naive.get(),
                fn_boundary_mask_naive.get(),
            )

        dilated_fp_boundary_mask = dilate_mask(
            mask=fp_boundary_mask_naive,
            width=boundary_width,
            implementation=self.boundary_implementation,
        )
        dilated_fn_boundary_mask = dilate_mask(
            mask=fn_boundary_mask_naive,
            width=boundary_width,
            implementation=self.boundary_implementation,
        )

        if GPU:
            dilated_fp_boundary_mask = np.asarray(dilated_fp_boundary_mask)
            dilated_fn_boundary_mask = np.asarray(dilated_fn_boundary_mask)

        fp_boundary_mask = np.logical_and(dilated_fp_boundary_mask, fp_mask)
        fn_boundary_mask = np.logical_and(dilated_fn_boundary_mask, fn_mask)

        if GPU:
            fp_boundary_mask = fp_boundary_mask.get()
            fn_boundary_mask = fn_boundary_mask.get()

        # check if every segment of boundary errors has a TP and a TN as direct neighbor
        fp_boundary_segments = get_contiguous_segments(fp_boundary_mask)
        fn_boundary_segments = get_contiguous_segments(fn_boundary_mask)

        tp_contour = get_exterior_boundary(tp_mask, width=1, implementation="fast")
        tn_contour = get_exterior_boundary(tn_mask, width=1, implementation="fast")

        for c, boundary_segments in fp_boundary_segments.items():
            for segment in boundary_segments:
                if (not tp_contour[c][segment].any()) or (not tn_contour[c][segment].any()):
                    fp_boundary_mask[c][segment] = False

        for c, boundary_segments in fn_boundary_segments.items():
            for segment in boundary_segments:
                if (not tp_contour[c][segment].any()) or (not tn_contour[c][segment].any()):
                    fn_boundary_mask[c][segment] = False

        results_on_mask = results[fp_boundary_mask]
        results[fp_boundary_mask] = np.where(
            results_on_mask != ERROR_CODES["unassigned"],
            results_on_mask,
            ERROR_CODES["FP_boundary"],
        )
        results_on_mask = results[fn_boundary_mask]
        results[fn_boundary_mask] = np.where(
            results_on_mask != ERROR_CODES["unassigned"],
            results_on_mask,
            ERROR_CODES["FN_boundary"],
        )
        return results

    def get_single_boundary_errors(self, class_results, tp_mask, tn_mask, fp_mask, fn_mask):
        H, W = tp_mask.shape[-2:]
        if self.use_relative_boundary_width:
            img_diag = np.sqrt(H**2 + W**2)
            if GPU:
                img_diag = img_diag.get()
                tp_mask = tp_mask.get()
                tn_mask = tn_mask.get()

            boundary_width = int(round(self.boundary_width * img_diag))
        else:
            boundary_width = self.boundary_width

        tp_ext_boundary = get_exterior_boundary(
            tp_mask, width=boundary_width, implementation=self.boundary_implementation
        )
        tn_ext_boundary = get_exterior_boundary(
            tn_mask, width=boundary_width, implementation=self.boundary_implementation
        )

        if GPU:
            tp_ext_boundary, tn_ext_boundary = np.asarray(tp_ext_boundary), np.asarray(
                tn_ext_boundary
            )

        boundary_intersection = np.logical_and(tp_ext_boundary, tn_ext_boundary)
        fp_boundary_mask_naive = np.logical_and(fp_mask, boundary_intersection)
        fn_boundary_mask_naive = np.logical_and(fn_mask, boundary_intersection)

        if GPU:
            fp_boundary_mask_naive, fn_boundary_mask_naive = (
                fp_boundary_mask_naive.get(),
                fn_boundary_mask_naive.get(),
            )

        dilated_fp_boundary_mask = dilate_mask(
            mask=fp_boundary_mask_naive,
            width=boundary_width,
            implementation=self.boundary_implementation,
        )
        dilated_fn_boundary_mask = dilate_mask(
            mask=fn_boundary_mask_naive,
            width=boundary_width,
            implementation=self.boundary_implementation,
        )

        if GPU:
            dilated_fp_boundary_mask = np.asarray(dilated_fp_boundary_mask)
            dilated_fn_boundary_mask = np.asarray(dilated_fn_boundary_mask)

        fp_boundary_mask = np.logical_and(dilated_fp_boundary_mask, fp_mask)
        fn_boundary_mask = np.logical_and(dilated_fn_boundary_mask, fn_mask)

        if GPU:
            fp_boundary_mask = fp_boundary_mask.get()
            fn_boundary_mask = fn_boundary_mask.get()

        # check if every segment of boundary errors has a TP and a TN as direct neighbor
        fp_boundary_segments = get_single_contiguous_segment(fp_boundary_mask)
        fn_boundary_segments = get_single_contiguous_segment(fn_boundary_mask)

        tp_contour = get_exterior_boundary(tp_mask, width=1, implementation="fast")
        tn_contour = get_exterior_boundary(tn_mask, width=1, implementation="fast")

        for segment in fp_boundary_segments:
            if (not tp_contour[segment].any()) or (not tn_contour[segment].any()):
                fp_boundary_mask[segment] = False

        for segment in fn_boundary_segments:
            if (not tp_contour[segment].any()) or (not tn_contour[segment].any()):
                fn_boundary_mask[segment] = False

        results_on_mask = class_results[fp_boundary_mask]
        class_results[fp_boundary_mask] = np.where(
            results_on_mask != ERROR_CODES["unassigned"],
            results_on_mask,
            ERROR_CODES["FP_boundary"],
        )
        results_on_mask = class_results[fn_boundary_mask]
        class_results[fn_boundary_mask] = np.where(
            results_on_mask != ERROR_CODES["unassigned"],
            results_on_mask,
            ERROR_CODES["FN_boundary"],
        )
        return class_results

    def get_extent_segment_errors(
        self,
        results,
        pred_one_hot,
        gt_one_hot,
    ):
        if GPU:
            pred_one_hot = pred_one_hot.get()
            gt_one_hot = gt_one_hot.get()

        pred_segments = get_contiguous_segments(pred_one_hot)
        gt_segments = get_contiguous_segments(gt_one_hot)

        for c, (pred_c, gt_c) in enumerate(zip(pred_one_hot, gt_one_hot)):
            if pred_c.any():
                if gt_c.any():
                    # positive
                    for pred_segment in pred_segments[c]:
                        results_on_segment = results[c][pred_segment]
                        if (results_on_segment == ERROR_CODES["unassigned"]).any():
                            error_type = (
                                "FP_extent"
                                if (results_on_segment == ERROR_CODES["TP"]).any()
                                else "FP_segment"
                            )
                            results[c][pred_segment] = np.where(
                                results_on_segment != ERROR_CODES["unassigned"],
                                results_on_segment,
                                ERROR_CODES[error_type],
                            )

                    # negative
                    for gt_segment in gt_segments[c]:
                        results_on_segment = results[c][gt_segment]
                        if (results_on_segment == ERROR_CODES["unassigned"]).any():
                            error_type = (
                                "FN_extent"
                                if (results_on_segment == ERROR_CODES["TP"]).any()
                                else "FN_segment"
                            )
                            results[c][gt_segment] = np.where(
                                results_on_segment != ERROR_CODES["unassigned"],
                                results_on_segment,
                                ERROR_CODES[error_type],
                            )
                else:  # only FP segment errors for this class
                    # positive prediction must be a superset of unassigned
                    # every prediction can only be unassigned or ignore
                    if GPU:
                        pred_c = np.asarray(pred_c)
                    assert pred_c[results[c] == ERROR_CODES["unassigned"]].all()
                    results[c][results[c] == ERROR_CODES["unassigned"]] = ERROR_CODES["FP_segment"]
            else:
                if gt_c.any():  # only FN segment errors for this class
                    results[c][results[c] == ERROR_CODES["unassigned"]] = ERROR_CODES["FN_segment"]
                else:
                    continue

        return results

    def get_single_extent_segment_errors(
        self,
        class_results,
        pred_mask,
        gt_mask,
    ):
        if GPU:
            pred_mask = pred_mask.get()
            gt_mask = gt_mask.get()

        pred_segments = get_single_contiguous_segment(pred_mask)
        gt_segments = get_single_contiguous_segment(gt_mask)

        if pred_mask.any():
            if gt_mask.any():
                # positive
                for pred_segment in pred_segments:
                    results_on_segment = class_results[pred_segment]
                    if (results_on_segment == ERROR_CODES["unassigned"]).any():
                        error_type = (
                            "FP_extent"
                            if (results_on_segment == ERROR_CODES["TP"]).any()
                            else "FP_segment"
                        )
                        class_results[pred_segment] = np.where(
                            results_on_segment != ERROR_CODES["unassigned"],
                            results_on_segment,
                            ERROR_CODES[error_type],
                        )

                # negative
                for gt_segment in gt_segments:
                    results_on_segment = class_results[gt_segment]
                    if (results_on_segment == ERROR_CODES["unassigned"]).any():
                        error_type = (
                            "FN_extent"
                            if (results_on_segment == ERROR_CODES["TP"]).any()
                            else "FN_segment"
                        )
                        class_results[gt_segment] = np.where(
                            results_on_segment != ERROR_CODES["unassigned"],
                            results_on_segment,
                            ERROR_CODES[error_type],
                        )
            else:  # only FP segment errors for this class
                # positive prediction must be a superset of unassigned
                # every prediction can only be unassigned or ignore
                if GPU:
                    pred_mask = np.asarray(pred_mask)
                assert pred_mask[class_results == ERROR_CODES["unassigned"]].all()
                class_results[class_results == ERROR_CODES["unassigned"]] = ERROR_CODES[
                    "FP_segment"
                ]
        else:
            if gt_mask.any():  # only FN segment errors for this class
                class_results[class_results == ERROR_CODES["unassigned"]] = ERROR_CODES[
                    "FN_segment"
                ]
            else:
                return class_results

        return class_results

    def evaluate_sample_boundary_iou(
        self, sample_results, pred_one_hot, gt_one_hot, ignore_inds=None
    ):
        H, W = sample_results.shape[-2:]
        img_diag = np.sqrt(H**2 + W**2)

        if GPU:
            img_diag = img_diag.get()
            pred_one_hot = pred_one_hot.get()
            gt_one_hot = gt_one_hot.get()

        boundary_width = max(int(round(self.boundary_iou_d * img_diag)), 1)

        # BoundaryIoU uses "fast" boundary implementation, see https://github.com/bowenc0221/boundary-iou-api/blob/37d25586a677b043ed585f10e5c42d4e80176ea9/boundary_iou/utils/boundary_utils.py#L12
        pred_one_hot_int_boundary = get_interior_boundary(
            pred_one_hot, width=boundary_width, implementation="fast"
        )  # P_d ∩ P
        gt_one_hot_int_boundary = get_interior_boundary(
            gt_one_hot, width=boundary_width, implementation="fast"
        )  # G_d ∩ G
        gt_one_hot_ext_boundary = get_exterior_boundary(
            gt_one_hot, width=boundary_width, implementation="fast"
        )  # G_d - G

        if GPU:
            pred_one_hot_int_boundary = np.asarray(pred_one_hot_int_boundary)
            gt_one_hot_int_boundary = np.asarray(gt_one_hot_int_boundary)

        boundary_intersection = np.logical_and(pred_one_hot_int_boundary, gt_one_hot_int_boundary)
        boundary_union = np.logical_or(pred_one_hot_int_boundary, gt_one_hot_int_boundary)

        boundary_intersection_counts = boundary_intersection.sum(axis=(1, 2))
        boundary_union_counts = boundary_union.sum(axis=(1, 2))

        return (
            boundary_intersection_counts,
            boundary_union_counts,
        )

    def evaluate_single_sample_boundary_iou(
        self, class_results, pred_mask, gt_mask, ignore_inds=None
    ):
        H, W = class_results.shape[-2:]
        img_diag = np.sqrt(H**2 + W**2)

        if GPU:
            img_diag = img_diag.get()
            pred_mask = pred_mask.get()
            gt_mask = gt_mask.get()

        boundary_width = max(int(round(self.boundary_iou_d * img_diag)), 1)

        # BoundaryIoU uses "fast" boundary implementation, see https://github.com/bowenc0221/boundary-iou-api/blob/37d25586a677b043ed585f10e5c42d4e80176ea9/boundary_iou/utils/boundary_utils.py#L12
        pred_mask_int_boundary = get_interior_boundary(
            pred_mask, width=boundary_width, implementation="fast"
        )  # P_d ∩ P
        gt_mask_int_boundary = get_interior_boundary(
            gt_mask, width=boundary_width, implementation="fast"
        )  # G_d ∩ G
        gt_mask_ext_boundary = get_exterior_boundary(
            gt_mask, width=boundary_width, implementation="fast"
        )  # G_d - G

        if GPU:
            pred_mask_int_boundary = np.asarray(pred_mask_int_boundary)
            gt_mask_int_boundary = np.asarray(gt_mask_int_boundary)

        boundary_intersection = np.logical_and(pred_mask_int_boundary, gt_mask_int_boundary)
        boundary_union = np.logical_or(pred_mask_int_boundary, gt_mask_int_boundary)

        if ignore_inds:  # remove ignore pixels
            ignore_inds_y, ignore_inds_x = ignore_inds
            assert not gt_mask[ignore_inds_y, ignore_inds_x].any()
            boundary_intersection[ignore_inds_y, ignore_inds_x] = 0
            boundary_union[ignore_inds_y, ignore_inds_x] = 0

        boundary_intersection_counts = boundary_intersection.sum()
        boundary_union_counts = boundary_union.sum()

        return (
            boundary_intersection_counts,
            boundary_union_counts,
        )

    def calculate_error_metrics(self):
        dataframe = pd.DataFrame(index=self.class_names)
        for error_name, error_counts in self.results.items():
            if error_name == "unassigned":
                assert (error_counts == 0).all()
                continue
            dataframe[error_name] = error_counts

        dataframe["FP"] = (
            dataframe["FP_boundary"] + dataframe["FP_extent"] + dataframe["FP_segment"]
        )
        dataframe["FN"] = (
            dataframe["FN_boundary"] + dataframe["FN_extent"] + dataframe["FN_segment"]
        )
        dataframe["E_boundary"] = dataframe["FP_boundary"] + dataframe["FN_boundary"]
        dataframe["E_extent"] = dataframe["FP_extent"] + dataframe["FN_extent"]
        dataframe["E_segment"] = dataframe["FP_segment"] + dataframe["FN_segment"]

        union = dataframe["TP"] + dataframe["FP"] + dataframe["FN"]
        dataframe["IoU"] = dataframe["TP"] / union
        dataframe["precision"] = dataframe["TP"] / (dataframe["TP"] + dataframe["FP"])
        dataframe["recall"] = dataframe["TP"] / (dataframe["TP"] + dataframe["FN"])
        dataframe["F1_score"] = 2 / (1.0 / dataframe["precision"] + 1.0 / dataframe["recall"])

        dataframe["FP_boundary_oU"] = dataframe["FP_boundary"] / union
        dataframe["FN_boundary_oU"] = dataframe["FN_boundary"] / union
        dataframe["E_boundary_oU"] = dataframe["E_boundary"] / union

        dataframe["FP_extent_oU"] = dataframe["FP_extent"] / union
        dataframe["FN_extent_oU"] = dataframe["FN_extent"] / union
        dataframe["E_extent_oU"] = dataframe["E_extent"] / union

        dataframe["FP_segment_oU"] = dataframe["FP_segment"] / union
        dataframe["FN_segment_oU"] = dataframe["FN_segment"] / union
        dataframe["E_segment_oU"] = dataframe["E_segment"] / union

        dataframe["E_boundary_oU_renormed"] = dataframe["E_boundary"] / (
            dataframe["TP"] + dataframe["E_boundary"]
        )
        dataframe["E_extent_oU_renormed"] = dataframe["E_extent"] / (
            dataframe["TP"] + dataframe["E_boundary"] + dataframe["E_extent"]
        )
        dataframe["E_segment_oU_renormed"] = dataframe["E_segment_oU"]

        with np.errstate(invalid="ignore"):  # avoid warnings for zero-division
            # boundary IoU
            dataframe["boundary_IoU"] = (
                self.boundary_iou_intersection_counts / self.boundary_iou_union_counts
            )
            # aggregate classes
            dataframe.loc["mean"] = dataframe.mean(axis=0)

        # dataframe.to_csv(f"{evaluator.result_dir}/result_df.csv", index=True)

        return dataframe



from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class MMSegmentation(BaseTrainArtifacts):
    def __init__(self, team_id: int):
        super().__init__(team_id)

        self._app_name = "Train MMSegmentation"
        self._framework_name = "MMSegmentation"
        self._framework_folder = "/mmsegmentation"
        self._weights_folder = "checkpoints/data"
        self._task_type = "instance segmentation"
        self._weights_ext = ".pth"
        self._config_file = "config.py"
        self._pattern = re_compile(r"^/mmsegmentation/\d+_[^/]+/?$")
        self._available_task_types: List[str] = ["instance segmentation"]

    def get_task_id(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[2].split("_")[0]

    def get_project_name(self, artifacts_folder: str) -> str:
        return artifacts_folder.split("/")[2].split("_")[1]

    def get_task_type(self, artifacts_folder: str) -> str:
        return self._task_type

    def get_config_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder, self._config_file)

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)



from __future__ import annotations

import numpy as np

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class ConfidenceDistribution(DetectionVisMetric):
    MARKDOWN = "confidence_distribution"
    CHART = "confidence_distribution"

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN,
            "Confidence Distribution",
            self.vis_texts.markdown_confidence_distribution.format(
                self.vis_texts.definitions.true_positives,
                self.vis_texts.definitions.false_positives,
            ),
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(self.CHART, self._get_figure())

    def _get_figure(self):  # -> go.Figure:
        import plotly.graph_objects as go  # pylint: disable=import-error

        f1_optimal_conf, best_f1 = self.eval_result.mp.m_full.get_f1_optimal_conf()

        # Histogram of confidence scores (TP vs FP)
        scores_tp, scores_fp = self.eval_result.mp.m_full.scores_tp_and_fp()

        tp_y, tp_x = np.histogram(scores_tp, bins=40, range=[0, 1])
        fp_y, fp_x = np.histogram(scores_fp, bins=40, range=[0, 1])
        dx = (tp_x[1] - tp_x[0]) / 2

        fig = go.Figure()
        fig.add_trace(
            go.Histogram(
                x=scores_fp,
                name="FP",
                marker=dict(color="#dd3f3f"),
                opacity=0.5,
                xbins=dict(size=0.025, start=0.0, end=1.0),
                hovertemplate="Confidence Score: %{x:.2f}<br>Count: %{y}<extra></extra>",
            )
        )
        fig.add_trace(
            go.Histogram(
                x=scores_tp,
                name="TP",
                marker=dict(color="#1fb466"),
                opacity=0.5,
                xbins=dict(size=0.025, start=0.0, end=1.0),
                hovertemplate="Confidence Score: %{x:.2f}<br>Count: %{y:.2f}<extra></extra>",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=tp_x + dx,
                y=tp_y,
                mode="lines+markers",
                name="TP",
                line=dict(color="#1fb466", width=2),
                hovertemplate="Confidence Score: %{x:.2f}<br>Count: %{y:.2f}<extra></extra>",
            )
        )
        fig.add_trace(
            go.Scatter(
                x=fp_x + dx,
                y=fp_y,
                mode="lines+markers",
                name="FP",
                line=dict(color="#dd3f3f", width=2),
                hovertemplate="Confidence Score: %{x:.2f}<br>Count: %{y:.2f}<extra></extra>",
            )
        )

        if f1_optimal_conf is not None:

            # Best threshold
            fig.add_shape(
                type="line",
                x0=f1_optimal_conf,
                x1=f1_optimal_conf,
                y0=0,
                y1=tp_y.max() * 1.3,
                line=dict(color="orange", width=1, dash="dash"),
            )
            fig.add_annotation(
                x=f1_optimal_conf,
                y=tp_y.max() * 1.3,
                text=f"F1-optimal threshold: {f1_optimal_conf:.2f}",
                showarrow=False,
            )

            fig.update_layout(
                barmode="overlay",
                width=800,
                height=500,
            )
            fig.update_xaxes(title_text="Confidence Score", range=[0, 1])
            fig.update_yaxes(title_text="Count", range=[0, tp_y.max() * 1.3])
        return fig



import os
from copy import deepcopy
from typing import Callable

import numpy as np

import supervisely as sly
from supervisely.io.fs import silent_remove


def get_image_by_hash(hash, save_path, api: sly.Api):
    api.image.download_paths_by_hashes([hash], [save_path])
    base_image = sly.image.read(save_path)
    silent_remove(save_path)
    return base_image


def download_image_from_context(
    context: dict,
    api: sly.Api,
    output_dir: str,
    cache_load_img: Callable[[sly.Api, int], np.ndarray] = None,
    cache_load_frame: Callable[[sly.Api, int, int], np.ndarray] = None,
    cache_load_img_hash: Callable[[sly.Api, str], np.ndarray] = None,
):
    if "image_id" in context:
        if cache_load_img is not None:
            return cache_load_img(api, context["image_id"])
        return api.image.download_np(context["image_id"])
    elif "image_hash" in context:
        if cache_load_img_hash is not None:
            return cache_load_img_hash(api, context["image_hash"])
        img_path = os.path.join(output_dir, "base_image.png")
        return get_image_by_hash(context["image_hash"], img_path, api=api)
    elif "volume" in context:
        volume_id = context["volume"]["volume_id"]
        slice_index = context["volume"]["slice_index"]
        normal = context["volume"]["normal"]
        window_center = context["volume"]["window_center"]
        window_width = context["volume"]["window_width"]
        plane = sly.Plane.get_name(normal)
        return api.volume.download_slice_np(
            volume_id, slice_index, plane, window_center, window_width
        )
    elif "video" in context:
        if cache_load_frame is not None:
            return cache_load_frame(
                api,
                context["video"]["video_id"],
                context["video"]["frame_index"],
            )
        return api.video.frame.download_np(
            context["video"]["video_id"], context["video"]["frame_index"]
        )
    else:
        raise Exception("Project type is not supported")


def crop_image(crop, image_np):
    x1, y1 = crop[0]["x"], crop[0]["y"]
    x2, y2 = crop[1]["x"], crop[1]["y"]
    bbox = sly.Rectangle(y1, x1, y2, x2)
    img_crop = sly.image.crop(image_np, bbox)
    return img_crop


def transform_clicks_to_crop(crop, clicks: list):
    clicks = deepcopy(clicks)
    for click in clicks:
        click["x"] -= crop[0]["x"]
        click["y"] -= crop[0]["y"]
    return clicks


def validate_click_bounds(crop, clicks: list):
    x_max = crop[1]["x"] - crop[0]["x"]  # width
    y_max = crop[1]["y"] - crop[0]["y"]  # height
    for click in clicks:
        is_in_bbox = (
            click["x"] >= 0 and click["y"] >= 0 and click["x"] <= x_max and click["y"] <= y_max
        )
        if not is_in_bbox:
            return False
    return True


def format_bitmap(bitmap: sly.Bitmap, crop):
    bitmap_json = bitmap.to_json()["bitmap"]
    bitmap_origin = bitmap_json["origin"]
    bitmap_origin = {
        "x": crop[0]["x"] + bitmap_origin[0],
        "y": crop[0]["y"] + bitmap_origin[1],
    }
    bitmap_data = bitmap_json["data"]
    return bitmap_origin, bitmap_data


def get_hash_from_context(context: dict):
    if "image_id" in context:
        return str(context["image_id"])
    elif "image_hash" in context:
        return context["image_hash"]
    elif "volume" in context:
        volume_id = context["volume"]["volume_id"]
        slice_index = context["volume"]["slice_index"]
        normal = context["volume"]["normal"]
        window_center = context["volume"]["window_center"]
        window_width = context["volume"]["window_width"]
        plane = sly.Plane.get_name(normal)
        return "_".join(map(str, [volume_id, slice_index, plane, window_center, window_width]))
    elif "video" in context:
        return "_".join(map(str, [context["video"]["video_id"], context["video"]["frame_index"]]))
    else:
        raise Exception("Project type is not supported")


def download_init_mask(api: sly.Api, figure_id, image_id) -> sly.Bitmap:
    ann_json = api.annotation.download_json(image_id)
    labels = [label for label in ann_json["objects"] if label["id"] == figure_id]
    assert len(labels) > 0, f"Label with id {figure_id} not found in image {image_id}."
    label = labels[0]
    bitmap = sly.Bitmap.from_json(label)
    return bitmap


def bitmap_to_mask(bitmap: sly.Bitmap, h, w):
    mask = np.zeros((h, w), bool)
    bitmap.to_bbox().get_cropped_numpy_slice(mask)[:] = bitmap.data
    mask = (mask * 255).astype(np.uint8)
    return mask



# coding: utf-8

from typing import List, Optional

from supervisely._utils import batched
from supervisely.api.module_api import ApiField, ModuleApi
from supervisely.collection.key_indexed_collection import KeyIndexedCollection
from supervisely.task.progress import tqdm_sly
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.task.progress import tqdm_sly


class TagApi(ModuleApi):
    """"""

    _entity_id_field = None
    """"""
    _method_bulk_add = None
    """"""

    @staticmethod
    def info_sequence():
        """
        NamedTuple TagInfo information about Tag.

        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            info_sequence = api.video.tag.info_sequence()
        """

        return [
            ApiField.ID,
            ApiField.PROJECT_ID,
            ApiField.NAME,
            ApiField.SETTINGS,
            ApiField.COLOR,
            ApiField.CREATED_AT,
            ApiField.UPDATED_AT,
        ]

    @staticmethod
    def info_tuple_name():
        """
        Get string name of NamedTuple for class.

        :return: NamedTuple name.
        :rtype: :class:`str`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            tuple_name = api.video.tag.info_tuple_name()
            print(tuple_name) # TagInfo
        """

        return "TagInfo"

    def get_list(self, project_id: int, filters=None):
        """
        Get list of tags for a given project ID.

        :param project_id: :class:`Dataset<supervisely.project.project.Project>` ID in Supervisely.
        :type project_id: int
        :param filters: List of parameters to sort output tags. See: https://api.docs.supervisely.com/#tag/Advanced/paths/~1tags.list/get
        :type filters: List[Dict[str, str]], optional
        :return: List of the tags from the project with given id.
        :rtype: list
        """

        return self.get_list_all_pages(
            "tags.list", {ApiField.PROJECT_ID: project_id, "filter": filters or []}
        )

    def get_name_to_id_map(self, project_id: int):
        """
        Get dictionary with mapping tag name to tag ID for a given project ID.

        :param project_id: :class:`Dataset<supervisely.project.project.Project>` ID in Supervisely.
        :type project_id: int
        :return: Dictionary with mapping tag name to tag id for a given project ID.
        :rtype: dict
        """

        tags_info = self.get_list(project_id)
        return {tag_info.name: tag_info.id for tag_info in tags_info}

    def _tags_to_json(self, tags: KeyIndexedCollection, tag_name_id_map=None, project_id=None):
        """"""
        if tag_name_id_map is None and project_id is None:
            raise RuntimeError("Impossible to get ids for project tags")
        if tag_name_id_map is None:
            tag_name_id_map = self.get_name_to_id_map(project_id)
        tags_json = []
        tags_keys = []
        for tag in tags:
            tag_json = tag.to_json()
            tag_json[ApiField.TAG_ID] = tag_name_id_map[tag.name]
            tags_json.append(tag_json)
            tags_keys.append(tag.key())
        return tags_json, tags_keys

    def append_to_entity(
        self,
        entity_id: int,
        project_id: int,
        tags: KeyIndexedCollection,
        key_id_map: KeyIdMap = None,
    ):
        """
        Add tags to entity in project with given ID.

        :param entity_id: ID of the entity in Supervisely to add a tag to
        :type entity_id: int
        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags: Collection of tags
        :type tags: KeyIndexedCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap, optional
        :return: List of tags IDs
        :rtype: list
        """

        if len(tags) == 0:
            return []
        tags_json, tags_keys = self._tags_to_json(tags, project_id=project_id)
        ids = self._append_json(entity_id, tags_json)
        KeyIdMap.add_tags_to(key_id_map, tags_keys, ids)
        return ids

    def _append_json(self, entity_id, tags_json):
        """"""
        if self._method_bulk_add is None:
            raise RuntimeError("self._method_bulk_add is not defined in child class")
        if self._entity_id_field is None:
            raise RuntimeError("self._entity_id_field is not defined in child class")

        if len(tags_json) == 0:
            return []
        response = self._api.post(
            self._method_bulk_add, {self._entity_id_field: entity_id, ApiField.TAGS: tags_json}
        )
        ids = [obj[ApiField.ID] for obj in response.json()]
        return ids

    def append_to_objects(
        self, entity_id: int, project_id: int, objects: KeyIndexedCollection, key_id_map: KeyIdMap
    ):
        """
        Add Tags to Annotation Objects for a specific entity (image etc.).

        :param entity_id: ID of the entity in Supervisely to add a tag to its objects
        :type entity_id: int
        :param project_id: Project ID in Supervisely. Uses to get tag name to tag ID mapping.
        :type project_id: int
        :param objects: Collection of Annotation Objects.
        :type objects: KeyIndexedCollection
        :param key_id_map: KeyIdMap object.
        :type key_id_map: KeyIdMap
        :return: List of tags IDs
        :rtype: list
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            os.environ['SERVER_ADDRESS'] = 'https://app.supervisely.com'
            os.environ['API_TOKEN'] = 'Your Supervisely API Token'
            api = sly.Api.from_env()

            img_id = 19373170
            img_info = api.
        """

        tag_name_id_map = self.get_name_to_id_map(project_id)

        tags_to_add = []
        tags_keys = []
        for object in objects:
            obj_id = key_id_map.get_object_id(object.key())
            if obj_id is None:
                raise RuntimeError(
                    "Can not add tags to object: OBJECT_ID not found for key {}".format(
                        object.key()
                    )
                )
            tags_json, cur_tags_keys = self._tags_to_json(
                object.tags, tag_name_id_map=tag_name_id_map
            )
            for tag in tags_json:
                tag[ApiField.OBJECT_ID] = obj_id
                tags_to_add.append(tag)
            tags_keys.extend(cur_tags_keys)

        if len(tags_keys) != len(tags_to_add):
            raise RuntimeError("SDK error: len(tags_keys) != len(tags_to_add)")
        if len(tags_keys) == 0:
            return
        ids = self.append_to_objects_json(entity_id, tags_to_add)
        KeyIdMap.add_tags_to(key_id_map, tags_keys, ids)
        return ids

    def append_to_objects_json(self, entity_id: int, tags_json: dict) -> list:
        """
        Add Tags to Annotation Objects for specific entity (image etc.).

        :param entity_id: ID of the entity in Supervisely to add a tag to its objects
        :type entity_id: int
        :param tags_json: Collection of tags in JSON format
        :type tags_json: dict
        :return: List of tags IDs
        :rtype: list
        """

        if len(tags_json) == 0:
            return []
        response = self._api.post(
            "annotation-objects.tags.bulk.add",
            {ApiField.ENTITY_ID: entity_id, ApiField.TAGS: tags_json},
        )
        ids = [obj[ApiField.ID] for obj in response.json()]
        return ids

    def add_to_objects(
        self,
        project_id: int,
        tags_list: List[dict],
        batch_size: int = 100,
        log_progress: bool = False,
        progress: Optional[tqdm_sly] = None,
    ) -> List[dict]:
        """
        Add Tags to existing Annotation Figures.
        All figures must belong to entities of the same project.
        Applies only to images project.

        :param project_id: Project ID in Supervisely.
        :type project_id: int
        :param tags_list: List of tag object infos as dictionaries.
        :type tags_list: List[dict]
        :param batch_size: Number of tags to add in one request.
        :type batch_size: int
        :param log_progress: If True, will display a progress bar.
        :type log_progress: bool
        :param progress: Progress bar object to display progress.
        :type progress: Optional[tqdm_sly]
        :return: List of tags infos as dictionaries.
        :rtype: List[dict]

        Usage example:
        .. code-block:: python

            import supervisely as sly

            api = sly.Api(server_address, token)

            tag_list = [
                {
                    "tagId": 25926,
                    "figureId": 652959,
                    "value": None # value is optional for tag with type 'None'
                },
                {
                    "tagId": 25927,
                    "figureId": 652959,
                    "value": "v1"
                },
                {
                    "tagId": 25927,
                    "figureId": 652958,
                    "value": "v2"
                }
            ]
            response = api.image.tag.add_to_figures(12345, tag_list)

            print(response)
            # Output:
            #    [
            #        {
            #            "id": 80421101,
            #            "tagId": 25926,
            #            "figureId": 652959,
            #            "value": None
            #        },
            #        {
            #            "id": 80421102,
            #            "tagId": 25927,
            #            "figureId": 652959,
            #            "value": "v1"
            #        },
            #        {
            #            "id": 80421103,
            #            "tagId": 25927,
            #            "figureId": 652958,
            #            "value": "v2"
            #        }
            #    ]
        """
        if type(self) is not TagApi:
            raise NotImplementedError("This method is not available for classes except TagApi")

        if len(tags_list) == 0:
            return []

        if progress is not None:
            log_progress = False

        result = []
        if log_progress:
            progress = tqdm_sly(
                desc="Adding tags to figures",
                total=len(tags_list),
            )
        for batch in batched(tags_list, batch_size):
            data = {ApiField.PROJECT_ID: project_id, ApiField.TAGS: batch}
            response = self._api.post("figures.tags.bulk.add", data)
            result.extend(response.json())
            if progress is not None:
                progress.update(len(batch))
        return result
